"""Video acquisition module (:mod:`pymanip.video`)
==================================================

This module defines the :class:`~pymanip.video.Camera` abstract base class,
which implements common methods such as the live video preview, and higher
level simple methods to quickly set up a video recording. It also
defines common useful functions, and a simple extension of Numpy arrays to
hold metadata (such as frame timestamp).

.. autoclass:: Camera
   :members:
   :private-members:

.. autoclass:: CameraTimeout

.. autoclass:: MetadataArray
   :members:
   :private-members:
   :show-inheritance:

   .. attribute:: metadata

      dictionnary attribute containing user-defined key-value pairs

.. autofunction:: save_image

"""

import sys
import signal
import os
import gzip
import numpy as np 

try:
    import pyqtgraph as pg
    from pyqtgraph.Qt import QtCore, QtGui
    from PyQt5.QtWidgets import (
        QMainWindow,
        QWidget,
        QPushButton,
        QVBoxLayout,
        QHBoxLayout,
        QApplication,
        QLabel,
    )

    has_qtgraph = True
except ModuleNotFoundError:
    has_qtgraph = False
    print("pyqtgraph is not installed.")
import cv2
import h5py
import time
from progressbar import ProgressBar
import asyncio
from pymanip.asynctools import synchronize_function, synchronize_generator


class CameraTimeout(Exception):
    """This class defines a CameraTimeout exception.
    """

    pass


def save_image(im, ii, basename, zerofill, file_format, compression, compression_level,
               color_order=None):
    """This function is a simple general function to save an input image from the camera
    to disk.

    :param im: input image
    :type im: :class:`~pymanip.video.MetadataArray`
    :param ii: frame number
    :type ii: int
    :param basename: file basename
    :type basename: str
    :param zerofill: number of digits for the frame number
    :type zerofill: int
    :param file_format: image file format on disk. Possible values are: "raw", "npy", "npy.gz", "hdf5", "png", or a file extension that OpenCV imwrite supports
    :type file_format: str
    :param compression: the compression argument "gzip" or "lzf" to pass to :meth:`h5py.create_dataset` if file_format is "hdf5"
    :type compression: str
    :param compression_level: the png compression level passed to opencv for the "png" file format
    :type compression_level: int

    """
    if file_format == "raw":
        filename = ("{:}-{:0" + str(zerofill) + "d}.li16").format(basename, ii + 1)
        im.tofile(filename)
    elif file_format == "npy":
        filename = ("{:}-{:0" + str(zerofill) + "d}.npy").format(basename, ii + 1)
        np.save(filename, im)
    elif file_format == "npy.gz":
        filename = ("{:}-{:0" + str(zerofill) + "d}.npy.gz").format(basename, ii + 1)
        with gzip.open(filename, "wb") as f:
            np.save(f, im)
    elif file_format in ("hdf", "hdf5"):
        filename = ("{:}-{:0" + str(zerofill) + "d}.hdf5").format(basename, ii + 1)
        with h5py.File(filename, "w") as f:
            f.attrs["counter"] = im.metadata["counter"]
            f.attrs["timestamp"] = im.metadata["timestamp"].timestamp()
            if color_order is not None:
                f.attrs["color_order"] = color_order
            # compression='gzip' trop lent pour 30 fps
            # compression='lzf' presque bon mais un peu lent à 30 fps
            f.create_dataset("image", data=im, compression=compression)
    else:
        filename = ("{:}-{:0" + str(zerofill) + "d}.{:}").format(
            basename, ii + 1, file_format
        )
        if file_format == "png":
            params = (cv2.IMWRITE_PNG_COMPRESSION, compression_level)
        else:
            params = None
        if color_order == "RGB":
            im = cv2.cvtColor(im, cv2.COLOR_RGB2BGR)
        cv2.imwrite(filename, im, params)


class MetadataArray(np.ndarray):
    """This class extends Numpy array to allow for an additionnal metadata
    attribute.
    """

    def __new__(cls, input_array, metadata=None):
        obj = np.asarray(input_array).view(cls)
        obj.metadata = metadata
        return obj

    def __array_finalize__(self, obj):
        if obj is None:
            return
        self.metadata = getattr(obj, "metadata", None)


class Camera:
    """This class is the abstract base class for all other concrete camera classes.
    The concrete sub-classes *must* implement the following methods:

    - :meth:`acquisition_oneshot` method

    - :meth:`acquisition` and :meth:`acquisition_async` generator methods

    - :attr:`resolution`, :attr:`name` and :attr:`bitdepth` properties

    The concrete sub-classes will also probably have to override the constructor
    method, and the enter/exit context manager method, as well as common property getters
    and setters:

    - :meth:`set_exposure_time`

    - :meth:`set_trigger_mode`

    - :meth:`set_roi`

    - :meth:`set_frame_rate`

    It may also define specialized getters for the camera which support them:

    - :meth:`set_adc_operating_mode`: ADC operating mode

    - :meth:`set_pixel_rate`: pixel rate sensor readout (in Hz)

    - :meth:`set_delay_exposuretime`

    """

    def acquisition_oneshot(self):
        """This method must be implemented in the sub-classes.
        It starts the camera, grab one frame, stops the camera, and returns the frame. It is useful for testing
        purposes, or in cases where only one frame is desired between very long time delays. It takes no input parameters.
        Returns an "autonomous" array (the buffer is independant of the camera object).

        :return: frame
        :rtype: :class:`~pymanip.video.MetadataArray`

        """
        raise NotImplementedError()

    def acquisition(
        self,
        num=np.inf,
        timeout=1000,
        raw=False,
        initialising_cams=None,
        raise_on_timeout=True,
    ):
        """This generator method is the main method that sub-classes must implement, along with the asynchronous variant.
        It is used by all the other higher-level methods, and can also be used directly in user code.

        :param num: number of frames to acquire, defaults to float("inf").
        :type num: int, or float("inf"), optional
        :param timout: timeout for frame acquisition (in milliseconds)
        :type timeout: int, optional
        :param raw: if True, returns bytes from the camera without any conversion. Defaults to False.
        :type raw: bool, optional
        :param initialising_cams: None, or set of camera objects. This camera object will remove itself from this set, once it is ready to grab frames. Useful in the case of multi camera acquisitions, to determine when all cameras are ready to grab frames. Defaults to None.
        :type initialising_cams: set, optional
        :param raise_on_timeout: boolean indicating whether to actually raise an exception when timeout occurs
        :type raise_on_timeout: bool, optional

        It starts the camera, yields :obj:`num` images, and closes the camera.
        It can be aborted when sent a true-truth value object. It then cleanly
        stops the camera and finally yields True as a confirmation that the stop_signal has been caught before returning.
        Sub-classes must therefore reads the possible stop_signal when yielding the frame, and act accordingly.

        The :class:`~pymanip.video.MetadataArray` objects yielded by this generator use a shared memory buffer which may
        be overriden for the next frame, and which is no longer defined when the generator object is cleaned up.
        The users are responsible for copying the array, if they want a persistant copy.

        User-level code will use the generator in this manner:

        .. code-block:: python

            gen = cam.acquire()
            for frame in gen:

                # .. do something with frame ..

                if I_want_to_stop:
                    clean = gen.send(True)
                    if not clean:
                        print('Warning generator not cleaned')
                    # no need to break here because the gen will be automatically exhausted

        """
        yield from synchronize_generator(
            self.acquisition_async, num, timeout, raw, None, raise_on_timeout
        )

    async def acquisition_async(
        self,
        num=np.inf,
        timeout=1000,
        raw=False,
        initialising_cams=None,
        raise_on_timeout=True,
    ):
        """This asynchronous generator method is similar to the :meth:`~pymanip.video.Camera.acquisition` generator method,
        except asynchronous. So much so, that in the general case, the latter can be defined simply by yielding from this
        asynchronous generator (so that the code is written once for both use cases), i.e.

        .. code-block:: python

            from pymanip.asynctools import synchronize_generator

            def acquisition(
                self,
                num=np.inf,
                timeout=1000,
                raw=False,
                initialising_cams=None,
                raise_on_timeout=True,
            ):
                yield from synchronize_generator(
                    self.acquisition_async,
                    num,
                    timeout,
                    raw,
                    initialising_cams,
                    raise_on_timeout,
                )

        It starts the camera, yields :obj:`num` images, and closes the camera.
        It can stop yielding images by sending the generator object a true-truth value object. It then cleanly
        stops the camera and finally yields True as a confirmation that the stop_signal has been caught before returning.
        Sub-classes must therefore reads the possible stop_signal when yielding the frame, and act accordingly.

        The :class:`~pymanip.video.MetadataArray` objects yielded by this generator use a shared memory buffer which may
        be overriden for the next frame, and which is no longer defined when the generator object is cleaned up.
        The users are responsible for copying the array, if they want a persistant copy.

        The user API is similar, except with asynchronous calls, i.e.

        .. code-block:: python

            gen = cam.acquire_async()
            async for frame in gen:

                # .. do something with frame ..

                if I_want_to_stop:
                    clean = await gen.asend(True)
                    if not clean:
                        print('Warning generator not cleaned')
                    # no need to break here because the gen will be automatically exhausted


        """
        raise NotImplementedError()

    def __enter__(self):
        """Context manager enter method
        """
        return self

    def __exit__(self, type_, value, cb):
        """Context manager exit method
        """
        if hasattr(self, "preview_generator"):
            self.preview_generator = None

    def preview(self, backend="cv", slice_=None, zoom=0.5, rotate=0):
        """This methods starts and synchronously runs the live-preview GUI.

        :param backend: GUI library to use. Possible values: "cv" for OpenCV GUI, "qt" for PyQtGraph GUI.
        :type backend: str
        :param slice_: coordinate of the region of interest to show, defaults to None
        :type slice_: Iterable[int], optional
        :param zoom: zoom factor, defaults to 0.5
        :type zoom: float, optional
        :param rotate: image rotation angle, defaults to 0
        :type rotate: float, optional

        """

        if backend == "cv":
            self.preview_cv(slice_, zoom, rotate)
        elif backend == "qt":
            self.preview_qt(slice_, zoom, None, rotate)
        else:
            raise RuntimeError('Unknown backend "' + backend + '"')

    async def preview_async_cv(self, slice_, zoom, name, rotate=0):
        """This method starts and asynchronously runs the live-preview with OpenCV GUI.
        The params are identical to the :meth:`~pymanip.video.Camera.preview` method.
        """

        minimum = None
        maximum = None
        cv2.namedWindow(name)
        try:
            preview_generator = self.acquisition_async()
            async for im in preview_generator:
                # if minimum is None:
                if True:
                    minimum = np.min(im)
                    maximum = np.max(im)
                    # print('min, max:', minimum, maximum)
                maxint = np.iinfo(im.dtype).max
                if rotate == 90.0:
                    im = cv2.rotate(im, cv2.ROTATE_90_COUNTERCLOCKWISE)
                elif rotate == -90.0:
                    im = cv2.rotate(im, cv2.ROTATE_90_CLOCKWISE)
                try:
                    l, c = im.shape
                    color = False
                except ValueError:
                    l, c, ncomp = im.shape
                    color = True

                if slice_:
                    if color:
                        img = (maxint // (maximum - minimum)) * (
                            im[slice_[0] : slice_[1], slice_[2] : slice_[3], :] - minimum
                        )
                    else:
                        img = (maxint // (maximum - minimum)) * (
                            im[slice_[0] : slice_[1], slice_[2] : slice_[3]] - minimum
                        )

                else:
                    img = (maxint // (maximum - minimum)) * (im - minimum)
                img = cv2.resize(img, (int(c * zoom), int(l * zoom)))
                if color and self.color_order == "RGB":
                    # OpenCV works in BGR order.
                    img = cv2.cvtColor(img, cv2.COLOR_RGB2BGR)
                cv2.imshow(name, img)
                k = cv2.waitKey(1)
                if k in (0x1B, ord("s")):
                    clean = await preview_generator.asend(True)
                    if not clean:
                        print("Generator not cleaned")
                    break
                await asyncio.sleep(0.001)
        except KeyboardInterrupt:
            pass
        finally:
            cv2.destroyAllWindows()

    def preview_cv(self, slice_, zoom, rotate=0):
        """This method starts and synchronously runs the live-preview with OpenCV GUI.
        It is a wrapper around the :meth:`pymanip.video.Camera.preview_async_cv` method.
        The params are identical to the :meth:`~pymanip.video.Camera.preview` method.
        """
        return synchronize_function(
            self.preview_async_cv, slice_, zoom, name="Preview", rotate=rotate
        )

    def preview_exitHandler(self):
        """This method sends a stop signal to the camera acquisition generator of the
        live-preview GUI.
        """

        clean = self.preview_generator.send(True)
        if not clean:
            print("Generator not cleaned")

    def display_crosshair(self):
        """This method adds a centered crosshair for self-reflection to the live-preview
        window (qt backend only)
        """
        if self.crosshair_chkbox.isChecked():
            self.vLine = pg.InfiniteLine(
                pos=(self.camera.Width / 2, 0), angle=90, movable=False
            )
            self.hLine = pg.InfiniteLine(
                pos=(0, self.camera.Height / 2), angle=0, movable=False
            )
            self.image_view.addItem(self.vLine, ignoreBounds=True)
            self.image_view.addItem(self.hLine, ignoreBounds=True)
        else:
            self.image_view.removeItem(self.vLine)
            self.image_view.removeItem(self.hLine)

    def preview_qt(self, slice, zoom, app=None, rotate=0):
        """This methods starts and synchronously runs the live-preview with Qt GUI.
        The params are identical to the :meth:`~pymanip.video.Camera.preview` method.
        """
        if app:
            self.app = app
            just_started = False
        elif not hasattr(self, "app"):
            self.app = QtGui.QApplication([])
            self.app.aboutToQuit.connect(self.preview_exitHandler)
            just_started = True
        else:
            just_started = False

        # create window if it does not already exists
        if not hasattr(self, "window"):
            self.window = QtGui.QMainWindow()
            # self.window.resize(*self.resolution)
            self.window.resize(800, 600)
            self.window.setWindowTitle(self.name)
            self.image_view = pg.ImageView()
            self.window.setCentralWidget(self.image_view)
            self.range_set = False

            # adding widget for controlling the background subtraction
            # and a crosshair overlay
            self.central_widget = QWidget()
            self.tools_widget = QWidget()
            self.central_layout = QVBoxLayout(self.central_widget)
            self.tools_layout = QHBoxLayout(self.tools_widget)

            self.crosshair_chkbox = QtGui.QCheckBox("Crosshair", self.tools_widget)
            self.subtraction_chkbox = QtGui.QCheckBox(
                "Background subtraction", self.tools_widget
            )
            self.learning_label = QLabel(
                "Learning rate  [0, 1] :", parent=self.tools_widget
            )
            self.spnbx_learning = QtGui.QDoubleSpinBox(
                parent=self.tools_widget, value=0.05
            )
            self.spnbx_learning.setRange(0, 1)
            self.spnbx_learning.setSingleStep(0.01)
            self.spnbx_learning.setDecimals(3)
            self.acq_btn = QtGui.QPushButton("Acquisition", self.tools_widget)
            self.exposure_label = QLabel(
                "Exposure time (s) :", parent=self.tools_widget
            )
            self.spnbox_exposure = QtGui.QDoubleSpinBox(
                parent=self.tools_widget, value=0.001
            )
            self.spnbox_exposure.setRange(0.000033, 67.108895)
            self.spnbox_exposure.setSingleStep(0.0001)
            self.spnbox_exposure.setDecimals(4)

            self.tools_layout.addWidget(self.crosshair_chkbox)
            self.tools_layout.addWidget(self.subtraction_chkbox)
            self.tools_layout.addWidget(self.learning_label)
            self.tools_layout.addWidget(self.spnbx_learning)
            self.tools_layout.addWidget(self.exposure_label)
            self.tools_layout.addWidget(self.spnbox_exposure)
            self.tools_layout.addWidget(self.acq_btn)

            self.central_layout.addWidget(self.image_view)
            self.central_layout.addWidget(self.tools_widget)

            self.window.setCentralWidget(self.central_widget)

            self.crosshair_chkbox.stateChanged.connect(self.display_crosshair)

            # hide useless buttons
            self.image_view.ui.roiBtn.hide()
            self.image_view.ui.menuBtn.hide()

            self.window.show()

        # instantiate generator
        if not hasattr(self, "preview_generator"):
            self.preview_generator = self.acquisition(timeout=1000, raise_on_timeout=False)

        if just_started:
            self.bkgrd = None

        # update view with latest image if it is ready
        # do nothing otherwise (to allow GUI interaction while waiting
        # for camera reading)
        img = next(self.preview_generator)
        if img is not None:
            try:
                l, c = img.shape
                color = False
            except ValueError:
                l, c, ncomp = img.shape
                color = True
            if zoom != 1.0:
                img = cv2.resize(img, (int(c * zoom), int(l * zoom)))
            if rotate == 90.0:
                img = cv2.rotate(img, cv2.ROTATE_90_COUNTERCLOCKWISE)
            elif rotate == -90.0:
                img = cv2.rotate(img, cv2.ROTATE_90_CLOCKWISE)

            if self.subtraction_chkbox.isChecked():
                if self.bkgrd is None:
                    self.bkgrd = img
                    self.range_set = False
                learning_rate = self.spnbx_learning.value()
                self.bkgrd = (1 - learning_rate) * self.bkgrd + learning_rate * img

                self.bkgrd = self.bkgrd.astype(np.int32)
                img = img - self.bkgrd  # self.bkgrd - img
                img[img < 0] = 0
                img = img.astype(np.uint16)
            if not color:
                img = img.T
            else:
                img = np.transpose(img, axes=(1,0,2))
                if self.color_order == "BGR":
                    # Qt comme Matplotlib travaille en RGB.
                    img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
            
            self.image_view.setImage(
                img, autoRange=False, autoLevels=False, autoHistogramRange=False
            )
            if not self.range_set:
                self.image_view.autoRange()
                self.image_view.autoLevels()
                self.range_set = True

        # set timer for refreshing in 10 ms
        QtCore.QTimer.singleShot(
            10, lambda: self.preview_qt(slice, zoom, self.app, rotate)
        )

        if just_started:
            QtGui.QApplication.instance().exec_()

    def acquire_to_files(self, *args, **kwargs):
        """This method starts the camera, acquires images and saves them to the disk.
        It is a simple wrapper around the :meth:`pymanip.video.Camera.acquire_to_files_async` asynchronous
        method. The parameters are identical.
        """
        return synchronize_function(self.acquire_to_files_async, *args, **kwargs)

    def acquire_signalHandler(self, *args, **kwargs):
        """This method sends a stop signal to the :meth:`~pymanip.video.Camera.acquire_to_files_async` method.
        """
        self.acqinterrupted = True

    async def acquire_to_files_async(
        self,
        num,
        basename,
        zerofill=4,
        dryrun=False,
        file_format="png",
        compression=None,
        compression_level=3,
        verbose=True,
        delay_save=False,
        progressbar=True,
        initialising_cams=None,
        **kwargs
    ):
        """This asynchronous method starts the camera, acquires :obj:`num` images and saves them to the disk. It is
        a simple quick way to perform camera acquisition (one-liner in the user code).

        :param num: number of frames to acquire
        :type num: int
        :param basename: basename for image filenames to be saved on disk
        :type basename: str
        :param zerofill: number of digits for the framenumber for image filename, defaults to 4
        :type zerofill: int, optional
        :param dryrun: do the acquisition, but saves nothing (testing purposes), defaults to False
        :type dryrun: bool, optional
        :param file_format: format for the image files, defaults to "png". Possible values are "raw", "npy", "npy.gz", "hdf5", "png" or any other extension supported by OpenCV imwrite.
        :type file_format: str, optional
        :param compression: compression option for HDF5 format ("gzip", "lzf"), defaults to None.
        :type compression: str, optional
        :param compression_level: png compression level for PNG format, defaults to 3.
        :type compression_level: int, optional
        :param verbose: prints information message, defaults to True.
        :type verbose: bool, optional
        :param delay_save: records all the frame in RAM, and saves at the end. This is useful for fast framerates when saving time is too slow. Defaults to False.
        :type delay_save: bool, optional
        :param progressbar: use :mod:`progressbar` module to show a progress bar. Defaults to True.
        :type progressbar: bool, optional
        :param initialising_cams: None, or set of camera objects. This camera object will remove itself from this set, once it is ready to grab frames. Useful in the case of multi camera acquisitions, to determine when all cameras are ready to grab frames. Defaults to None.
        :type initialising_cams: set, optional
        :return: image_counter, frame_datetime
        :rtype: list, list

        The details of the file format are given in this table:

        ===============     ========================================================================
        file_format         description
        ===============     ========================================================================
        raw                 native 16 bits integers, i.e. li16 (little-endian) on Intel CPUs
        npy                 numpy npy file (warning: depends on pickle format)
        npy.gz              gzip compressed numpy file
        hdf5                hdf5, with optional compression
        png, jpg, tif       image format with opencv imwrite with optional compression level for PNG
        ===============     ========================================================================

        Typical usage of the function for one camera:

        .. code-block:: python

            async def main():

                with Camera() as cam:
                    counts, times = await cam.acquire_to_files_async(num=20, basename='img-')

            asyncio.run(main())


        """

        # signal handling
        if sys.platform == "win32":
            signal.signal(signal.SIGINT, self.acquire_signalHandler)
        else:
            loop = asyncio.get_event_loop()
            for signame in ("SIGINT", "SIGTERM"):
                loop.add_signal_handler(
                    getattr(signal, signame), self.acquire_signalHandler
                )

        dirname = os.path.dirname(basename)
        if len(dirname):
            try:
                os.makedirs(dirname)
            except FileExistsError:
                pass
        count = []
        dt = []
        if verbose:
            dateformat = "%A %d %B %Y - %X"
            starttime = time.time()
            starttime_str = time.strftime(dateformat, time.localtime(starttime))
            print("Camera acquisition started: " + starttime_str)
        if progressbar:
            bar = ProgressBar(max_value=num)
        computation_time = 0.0
        images = list()
        ii = 0
        acqgen = self.acquisition_async(
            num, initialising_cams=initialising_cams, **kwargs
        )
        self.acqinterrupted = False
        async for im in acqgen:
            if ii == 0:
                print(im.dtype)
            if not dryrun:
                if delay_save:
                    images.append(im.copy())
                else:
                    start_time = time.process_time()
                    if hasattr(self, "color_order"):
                        color_order = self.color_order
                    else:
                        color_order = None
                    if im.ndim < 3:
                        color_order = None
                    save_image(
                        im,
                        ii,
                        basename,
                        zerofill,
                        file_format,
                        compression,
                        compression_level,
                        color_order=color_order,
                    )
                    computation_time += time.process_time() - start_time
            if hasattr(im, "metadata"):
                count.append(im.metadata["counter"])
                ts = im.metadata["timestamp"]
                try:
                    ts = ts.timestamp()
                except AttributeError:
                    pass
                dt.append(ts)
            ii += 1
            if progressbar:
                try:
                    bar.update(ii)
                except Exception:
                    print(ii)
            await asyncio.sleep(0.001)
            if self.acqinterrupted:
                print("")
                print("Signal caught... Stopping camera acquisition...")
                clean = await acqgen.asend(True)
                if not clean:
                    print("Camera was not successfully interrupted")
                break
        if progressbar:
            print("")
        if delay_save and not dryrun:
            print("Acquisition complete. Saving to disk...")
            if progressbar:
                bar = ProgressBar(max_value=ii)
            for ii, im in enumerate(images):
                start_time = time.process_time()
                save_image(
                    im,
                    ii,
                    basename,
                    zerofill,
                    file_format,
                    compression,
                    compression_level,
                )
                computation_time += time.process_time() - start_time
                if progressbar:
                    try:
                        bar.update(ii + 1)
                    except Exception:
                        print(ii)
                await asyncio.sleep(0.0001)
        if progressbar:
            print("")
        dt = np.array(dt)
        if verbose:
            print(
                "Average saving time per image:",
                1000 * computation_time / (ii + 1),
                "ms",
            )
            print("average fps =", 1 / np.mean(dt[1:] - dt[:-1]))
            if images:
                print("image size:", images[0].shape)
                print("image dtype:", images[0].dtype)
        return count, dt



#! /usr/bin/env python
# -*- coding: utf-8 -*-

# from distutils.core import setup
from setuptools import setup, find_packages


def readme():
    with open("README.md", encoding="utf8") as f:
        return f.read()


setup(
    name="pymanip",
    version="0.4",
    description="Python framework for experiments",
    long_description=readme(),
    long_description_content_type="text/markdown",
    url="https://pypi.org/project/pymanip/",
    project_urls={
        "Documentation": "https://pymanip.readthedocs.io/en/latest/",
        "Source Code": "https://github.com/jsalort/pymanip",
    },
    author="Julien Salort",
    author_email="julien.salort@ens-lyon.fr",
    license="CeCILL-B",
    packages=find_packages(),
    install_requires=[
        "h5py",
        "fluiddyn >= 0.4.0",
        "fluidlab >=0.1.0",
        "progressbar2",
        "aiohttp",
        "aiohttp_jinja2",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
"""
Module for experimental sessions.

Useful classes are Session and SavedSession.
"""

import os
import numpy as np
import time
import inspect
import matplotlib.pyplot as plt
import warnings

with warnings.catch_warnings():
    warnings.filterwarnings("ignore", category=FutureWarning)
    warnings.filterwarnings("ignore", category=DeprecationWarning)
    import h5py
import smtplib
import base64
import quopri
import tempfile

from datetime import datetime

from fluiddyn.util.terminal_colors import cprint

try:
    from convbox.myplot import ColorGenerator
except ImportError:
    import itertools

    def ColorGenerator():
        return itertools.cycle(["b", "r", "g", "k", "m", "c"])


import pymanip.mytime as mytime

try:
    from pathlib import Path

    has_pathlib = True
except ImportError:
    has_pathlib = False
from pymanip.mytime import dateformat

__all__ = ["makeAcqName", "SavedSession", "Session", "NameGenerator"]


def NameGenerator(prefix=None, postfix=None):
    acquisition_clock = datetime.now()
    acquisition_number = 1
    name = "%d-%02d-%02d_%02d-%02d-%02d" % (
        acquisition_clock.year,
        acquisition_clock.month,
        acquisition_clock.day,
        acquisition_clock.hour,
        acquisition_clock.minute,
        acquisition_clock.second,
    )
    if prefix:
        name = prefix + "_" + name
    if postfix:
        name = name + "_" + postfix
    while True:
        name_numbered = name + "_" + str(acquisition_number)
        print("Acquisition name:", name_numbered)
        yield name_numbered
        acquisition_number = acquisition_number + 1


defaultGenerator = None


def makeAcqName(comment=None):
    global defaultGenerator

    if (
        (defaultGenerator is None)
        or (comment == "reset")
        or (defaultGenerator.gi_frame.f_locals["postfix"] != comment)
    ):
        defaultGenerator = NameGenerator(postfix=comment)
    return next(defaultGenerator)


class BaseSession(object):
    def __init__(self, session_name=None):
        if session_name is None:
            session_name = makeAcqName()

        self.session_name = session_name
        self.storename = session_name + ".hdf5"
        self.session_opening_time = time.time()
        self.opened = False
        self.parameters_defined = False
        self.grp_datasets_defined = False
        self.allow_override_datasets = False

    def __str__(self):
        return self.session_name

    def describe(self):
        # Logged variables
        if len(self.grp_variables.keys()) > 0:
            num_lines = self.dset_time.len()
            print("List of saved variables: (%d lines)" % num_lines)
            for var in self.grp_variables.keys():
                print(" " + var)
        # Datasets
        if self.grp_datasets_defined:
            print("List of saved datasets:")
            for dataname in self.grp_datasets.keys():
                size = self.grp_datasets[dataname].size
                print(" " + dataname + (" (%d points)" % size))
        # Parameters
        if self.parameters_defined:
            if len(self.parameters.keys()) > 0:
                print("List of saved parameters:")
                for name in self.parameters.keys():
                    value = self.parameters[name]
                    # print(type(value))
                    if isinstance(value, np.ndarray) and len(value) == 1:
                        value = value[0]
                    if name == "email_lastSent":
                        theDateStr = time.strftime(dateformat, time.localtime(value))
                        print(" " + name + " = " + theDateStr)
                    else:
                        print(
                            " "
                            + name
                            + " = "
                            + str(value)
                            + " ("
                            + str(type(value))
                            + ")"
                        )

    def has_dataset(self, name):
        if self.grp_datasets_defined:
            return name in self.grp_datasets.keys()
        else:
            return False

    def dataset(self, name):
        if self.grp_datasets_defined:
            return self.grp_datasets[name][()]

    def dataset_names(self):
        if hasattr(self, "grp_datasets"):
            return self.grp_datasets.keys()
        return []

    def has_parameter(self, name):
        if self.parameters_defined:
            return name in self.parameters.keys()
        else:
            return False

    def parameter(self, name):
        if self.parameters_defined:
            return self.parameters[name]

    def has_log(self, name):
        if name == "Time" or name == "time" or name == "t":
            return True
        return name in self.grp_variables.keys()

    def log_variable_list(self):
        return self.grp_variables.keys()

    def log(self, varname):
        if self.opened:
            if varname == "Time" or varname == "time" or varname == "t":
                return self.dset_time[()]
            elif varname == "?":
                print("List of saved variables:")
                for var in self.grp_variables.keys():
                    print(var)
            elif varname in self.grp_variables.keys():
                return self.grp_variables[varname][()]
            else:
                cprint.red("Variable is not defined: " + varname)
        else:
            cprint.red("Session is not opened")

    def __getitem__(self, key):
        if self.has_log(key):
            return self.log(key)
        elif self.has_parameter(key):
            return self.parameter(key)
        elif self.has_dataset(key):
            return self.dataset(key)
        else:
            raise KeyError('Unknown key "' + key + '"')

    def log_plot(self, fignum, varlist, maxvalues=1000, plotfunc=plt.plot):
        if self.opened:
            plt.figure(fignum)
            plt.clf()
            plt.ion()
            plt.show()
            t = self.log("t")
            if len(t) > maxvalues:
                debut = len(t) - 1000
                fin = len(t)
            else:
                debut = 0
                fin = len(t)
            if t[debut] > self.session_opening_time:
                # tous les points sont nouveaux
                olddebut = None
                oldfin = None
                newdebut = debut
                newfin = fin
            elif t[-1] > self.session_opening_time:
                # certains points sont nouveaux
                bb = t > self.session_opening_time
                olddebut = debut
                oldfin = np.min(bb.argmax())
                newdebut = oldfin
                newfin = fin
            else:
                # les points sont tous anciens
                olddebut = debut
                oldfin = fin
                newdebut = None
                newfin = None
            if t[-1] - t[0] < 3600:
                t = (t - t[0]) / 60.0
                xlab = "t [min]"
            else:
                t = (t - t[0]) / 3600.0
                xlab = "t [h]"
            # print(olddebut, oldfin, newdebut, newfin)

            if isinstance(varlist, str):
                lab = varlist
                col = (0, 0, 1)
                if newdebut is not None:
                    plt.plot(
                        t[newdebut:newfin],
                        self.log(varlist)[newdebut:newfin],
                        "o-",
                        color=col,
                        mec=col,
                        mfc=col,
                        label=lab,
                    )
                    lab = None
                if olddebut is not None:
                    plt.plot(
                        t[olddebut:oldfin],
                        self.log(varlist)[olddebut:oldfin],
                        "o-",
                        mfc="none",
                        mec=col,
                        color=col,
                        label=lab,
                    )

            else:
                for var, coul in zip(varlist, ColorGenerator()):
                    lab = var
                    if newdebut is not None:
                        plotfunc(
                            t[newdebut:newfin],
                            self.log(var)[newdebut:newfin],
                            "o-",
                            mfc=coul,
                            mec=coul,
                            color=coul,
                            label=lab,
                        )
                        lab = None
                    if olddebut is not None:
                        plotfunc(
                            t[olddebut:oldfin],
                            self.log(var)[olddebut:oldfin],
                            "o-",
                            mfc="none",
                            mec=coul,
                            color=coul,
                            label=lab,
                        )

            plt.xlabel(xlab)
            plt.legend(loc="upper left")
            plt.draw()
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                plt.pause(0.0001)
        else:
            cprint.red("Session is not opened")

    def sleep(self, duration):
        mytime.sleep(duration)

    def __enter__(self):
        self.exited = False
        return self

    def __exit__(self, type_, value, cb):
        self.exited = True


class SavedSession(BaseSession):
    def __init__(
        self, session_name, cache_override=False, cache_location=".", verbose=True
    ):
        super(SavedSession, self).__init__(session_name)
        self.store = h5py.File(self.storename, "r")
        try:
            self.dset_time = self.store["time"]
        except KeyError:
            print("The file '" + self.storename + "' is not a pymanip session file.")
            raise RuntimeError("Wrong hdf5 data")
        self.grp_variables = self.store["variables"]
        self.verbose = verbose
        try:
            self.parameters = self.store.attrs
            self.parameters_defined = True
        except Exception:
            self.parameters_defined = False
            pass
        try:
            self.grp_datasets = self.store["datasets"]
            self.grp_datasets_defined = True
        except Exception:
            self.grp_datasets_defined = False
            pass
        self.opened = True
        if verbose:
            print("Loading saved session from file", self.storename)
        total_size = self.dset_time.len()
        if total_size > 0:
            start_t = self.dset_time[0]
            end_t = self.dset_time[total_size - 1]
            start_string = time.strftime(dateformat, time.localtime(start_t))
            end_string = time.strftime(dateformat, time.localtime(end_t))
            if verbose:
                cprint.blue("*** Start date: " + start_string)
                cprint.blue("***   End date: " + end_string)
        elif not self.grp_datasets_defined:
            if verbose:
                cprint.red("No logged variables")
        if self.grp_datasets_defined:
            timestamp_string = time.strftime(
                dateformat, time.localtime(self.grp_datasets.attrs["timestamp"])
            )
            if verbose:
                cprint.blue("*** Acquisition timestamp " + timestamp_string)
        self.cachestorename = os.path.join(
            os.path.realpath(cache_location), "cache", os.path.basename(self.storename)
        )
        if cache_override:
            self.cachemode = "w"
        else:
            self.cachemode = "r+"
        try:
            self.cachestore = h5py.File(self.cachestorename, self.cachemode)
            if verbose:
                cprint.yellow("*** Cache store found at " + self.cachestorename)
            self.has_cachestore = True
        except IOError:
            self.has_cachestore = False
            pass

    @property
    def cachedvars(self):
        if not hasattr(self, "has_cachestore"):
            return []
        if self.has_cachestore:
            return self.cachestore.keys()
        else:
            return []

    def cached(self, *args):
        """
        cached('var1', 'var2', ...)
        or
        cached(['var1', 'var2'])

        returns True if all specified variables are present in the cachestore
        """
        if len(args) == 1:
            name = args[0]
        else:
            name = [a for a in args]
        if isinstance(name, str):
            return name in self.cachedvars
        else:
            return all([(n in self.cachedvars) for n in name])

    def cachedvalue(self, varname):
        if self.has_cachestore:
            if self.verbose:
                cprint.yellow("Retrieving " + varname + " from cache")
            content = self.cachestore[varname]
            if hasattr(content, "value"):
                return content[()]
            else:
                result = list()
                i = 0
                while True:
                    try:
                        result.append(content[str(i)][()])
                        i = i + 1
                    except KeyError:
                        break
                return np.array(result)
        else:
            raise KeyError

    def cache(self, name, dict_caller=None):
        if dict_caller is None:
            stack = inspect.stack()
            try:
                dict_caller = stack[1][0].f_locals
            finally:
                del stack
        if not isinstance(name, str):
            for var in name:
                self.cache(var, dict_caller)
            return

        if not self.has_cachestore:
            try:
                os.mkdir(os.path.dirname(self.cachestorename))
            except OSError:
                pass
            try:
                self.cachestore = h5py.File(self.cachestorename, "w")
                self.has_cachestore = True
                cprint.yellow("*** Cache store created at " + self.cachestorename)
            except IOError as ioe:
                self.has_cachestore = False
                cprint.red("Cannot create cache store")
                cprint.red(ioe.message)

        if self.has_cachestore:
            cprint.yellow("Saving " + name + " in cache")

            if isinstance(dict_caller[name], list):
                # Sauvegarde d'une liste d'objets
                # Dans ce cas, on crée un groupe du nom "name" et on récurse
                # Si le groupe existe déjà, on le supprime et on le recrée
                try:
                    del self.cachestore[name]
                except KeyError:
                    pass
                grp = self.cachestore.create_group(name)
                for itemnum, item in enumerate(dict_caller[name]):
                    if not isinstance(item, (int, float)):
                        grp.create_dataset(
                            str(itemnum), chunks=True, data=item, compression="gzip"
                        )
                    else:
                        grp.create_dataset(str(itemnum), data=item)
            else:

                try:
                    new_length = len(dict_caller[name])
                except TypeError:
                    new_length = 1
                    pass
                if name in self.cachestore.keys():
                    if len(self.cachestore[name]) != new_length:
                        self.cachestore[name].resize((new_length,))
                    self.cachestore[name][:] = dict_caller[name]
                else:
                    if new_length > 1:
                        self.cachestore.create_dataset(
                            name, chunks=True, maxshape=(None,), data=dict_caller[name]
                        )
                    else:
                        self.cachestore.create_dataset(
                            name, chunks=True, maxshape=(None,), shape=(new_length,)
                        )
                        self.cachestore[name][:] = dict_caller[name]

    def __exit__(self, type, value, cb):
        """ Previous versions used __del__ which is bad
            practise because the gc triggering mechanism is
            an implementation detail of CPython.
            Therefore, one should either explicitly call
            Stop() or close(), or use a context manager
            (preferred way)
        """
        self.store.close()
        self.store = None
        self.opened = False
        if hasattr(self, "has_cachestore") and self.has_cachestore and self.cachestore:
            self.cachestore.close()
            self.has_cachestore = False
            self.cachestore = None
        super(SavedSession, self).__exit__(type, value, cb)


class Session(BaseSession):
    def __init__(self, session_name, variable_list=[], allow_override_datasets=False):
        if has_pathlib:
            if isinstance(session_name, Path):
                session_name = session_name.as_posix()
        super(Session, self).__init__(session_name)
        self.datname = session_name + ".dat"
        self.logname = session_name + ".log"
        self.datfile = open(self.datname, "a")
        self.logfile = open(self.logname, "a")
        self.allow_override_datasets = allow_override_datasets

        date_string = time.strftime(
            dateformat, time.localtime(self.session_opening_time)
        )
        self.logfile.write("Session opened on " + date_string)
        self.logfile.flush()
        try:
            self.store = h5py.File(self.storename, "r+")
            self.dset_time = self.store["time"]
            self.grp_variables = self.store["variables"]
            self.parameters = self.store.attrs
            self.parameters_defined = True
            try:
                self.grp_datasets = self.store["datasets"]
                self.grp_datasets_defined = True
            except Exception:
                self.grp_datasets_defined = False
            original_size = self.dset_time.len()
            arr = np.zeros((original_size,))
            new_headers = False
            if len(variable_list) != len(self.grp_variables.keys()):
                new_headers = True
            for var in variable_list:
                if var not in self.grp_variables.keys():
                    self.grp_variables.create_dataset(
                        var, chunks=True, maxshape=(None,), data=arr
                    )
                    new_headers = True
            cprint.blue("Session reloaded from file", bold=True, end=" ")
            cprint.blue(self.storename)
            if original_size > 0:
                last_t = self.dset_time[original_size - 1]
                date_string = time.strftime(dateformat, time.localtime(last_t))
                cprint.black("Last point recorded:", bold=True, end=" ")
                print(date_string)
        except IOError:
            self.store = h5py.File(self.storename, "w")
            self.dset_time = self.store.create_dataset(
                "time", chunks=True, maxshape=(None,), shape=(0,), dtype=float
            )
            self.grp_variables = self.store.create_group("variables")
            self.parameters = self.store.attrs
            self.parameters_defined = True
            self.parameters["email_lastSent"] = 0.0
            new_headers = True
            for var in variable_list:
                self.grp_variables.create_dataset(
                    var, chunks=True, maxshape=(None,), shape=(0,), dtype=float
                )
        if new_headers:
            self.datfile.write("Time")
            # attention: ne pas utiliser variable_list ici car
            # dans log_addline on utilise self.grp_variable.keys()
            # et l'ordre n'est pas le même
            for var in self.grp_variables.keys():
                self.datfile.write(" " + var)
            self.datfile.write("\n")
        self.opened = True
        self.email_started = False

    def disp(self, texte):
        print(texte)
        if self.email_started:
            self.email_body = self.email_body + texte + "<br />\n"
        if not texte.endswith("\n"):
            texte = texte + "\n"
        self.logfile.write(texte)
        self.logfile.flush()

    def log_addline(self, timestamp=None, dict_caller=None):
        if not dict_caller:
            stack = inspect.stack()
            try:
                dict_caller = stack[1][0].f_locals
            finally:
                del stack
        newsize = self.dset_time.len() + 1
        self.dset_time.resize((newsize,))
        if not timestamp:
            timestamp = time.time()
        self.dset_time[newsize - 1] = timestamp
        self.datfile.write("%f" % self.dset_time[newsize - 1])
        for varname in self.grp_variables.keys():
            d = self.grp_variables[varname]
            d.resize((newsize,))
            try:
                d[newsize - 1] = dict_caller[varname]
                self.datfile.write(" %f" % dict_caller[varname])
            except Exception:
                cprint.red("Variable is not defined: " + varname)
                d[newsize - 1] = 0.0
        self.datfile.write("\n")
        self.datfile.flush()
        self.store.flush()

    def save_remote_data(self, data):
        """
        Save data from RemoteObserver object as datasets and parameters
        """
        for k, v in data.items():
            # print(k,type(v),v)
            try:
                v[0]
                # we are iterable
                self.save_dataset(k, data)
            except (TypeError, KeyError):
                # we are not iterable
                if isinstance(v, dict):
                    # non reduced data, v is a dictionnary with two keys, 't' and 'value'
                    self.save_dataset(k, {k: v["value"]})
                    self.save_dataset(k + "_time", {k + "_time": v["t"]})
                else:
                    try:
                        # data must be a scalar
                        float(v)
                    except TypeError:
                        print("skipping", k, type(v))
                        continue
                    self.save_parameter(k, data)

    def save_parameter(self, parameter_name, dict_caller=None):
        if dict_caller is None:
            stack = inspect.stack()
            try:
                dict_caller = stack[1][0].f_locals
            finally:
                del stack
        if isinstance(parameter_name, str):
            try:
                value = dict_caller[parameter_name]
                self.parameters[parameter_name] = value
            except TypeError:
                print(f"Failed to save parameter {parameter_name:}")
                print(type(value))
                print(value)
                raise
        elif isinstance(parameter_name, dict):
            for k, v in parameter_name.items():
                self.parameters[k] = v
        else:
            for var in parameter_name:
                value = dict_caller[var]
                self.parameters[var] = value

    def save_parameters(self, parameter_list):
        """
        save_parameters( parameter_list )
        or
        save_parameters( parameter_dict)
        """

        if isinstance(parameter_list, dict):
            for k, v in parameter_list.items():
                self.parameters[k] = v
        else:
            stack = inspect.stack()
            try:
                dict_caller = stack[1][0].f_locals
            finally:
                del stack
            self.save_parameter(parameter_list, dict_caller)

    def save_dataset(self, data_name, dict_caller=None):
        if dict_caller is None:
            stack = inspect.stack()
            try:
                dict_caller = stack[1][0].f_locals
            finally:
                del stack
        if not self.grp_datasets_defined:
            self.grp_datasets = self.store.create_group("datasets")
            self.grp_datasets_defined = True
        if data_name not in self.grp_datasets.keys():
            self.grp_datasets.attrs["timestamp"] = time.time()
            self.grp_datasets.create_dataset(
                data_name, chunks=True, maxshape=(None,), data=dict_caller[data_name]
            )
        elif self.allow_override_datasets:
            new_length = len(dict_caller[data_name])
            if len(self.grp_datasets[data_name]) != new_length:
                self.grp_datasets[data_name].resize((new_length,))
            self.grp_datasets[data_name][:] = dict_caller[data_name]
            self.grp_datasets.attrs["timestamp"] = time.time()
            cprint.red("Warning: overriding existing dataset")
        else:
            raise NameError(
                "Dataset is already defined. Use allow_override_datasets to allow override of existing saved datasets."
            )

    def save_datasets(self, data_list):
        stack = inspect.stack()
        try:
            dict_caller = stack[1][0].f_locals
        finally:
            del stack
        for data_name in data_list:
            self.save_dataset(data_name, dict_caller)

    def start_email(self, from_addr, to_addrs, host, subject=None, port=25):
        self.email_host = host
        self.email_port = port
        self.email_from_addr = from_addr
        self.email_to_addrs = to_addrs
        date_string = time.strftime(dateformat, time.localtime(time.time()))
        if subject is not None:
            self.email_subject = subject
        else:
            self.email_subject = self.session_name
        self.email_body = (
            "<html><body>\n<strong>**************************************************</strong><br />\n<strong>"
            + date_string
            + "</strong><br />\n"
            + self.session_name
            + "<br />\n<strong>**************************************************</strong><br /><br />\n\n"
        )
        self.email_figlist = []
        self.email_started = True

    def add_figure_to_email(self, figNum):
        self.email_figlist.append(figNum)

    def stop_email(self):
        success = False
        smtp = smtplib.SMTP(self.email_host, self.email_port)
        self.email_body = self.email_body + "</body></html>"

        useMime = False
        if len(self.email_figlist) > 0:
            useMime = True

        mime_boundary = "pymanip-MIME-delimiter"
        if useMime:
            email_header = (
                'Content-type: multipart/mixed; boundary="' + mime_boundary + '"\n'
            )
            email_header = email_header + "MIME-version: 1.0\n"
        else:
            email_header = "Content-Type: text/html\n"
            email_header = (
                email_header + "Content-Transfer-Encoding: quoted-printable\n"
            )
        email_header = email_header + "User-Agent: pymanip\n"
        email_header = email_header + "To: "
        if isinstance(self.email_to_addrs, str):
            email_header = email_header + self.email_to_addrs + "\n"
        elif isinstance(self.email_to_addrs, tuple):
            for addr in self.email_to_addrs[:-1]:
                email_header = email_header + addr + ", "
            email_header = email_header + self.email_to_addrs[-1] + "\n"
        else:
            raise ValueError("Adress list should be a string or a tuple")
        email_header = email_header + "Subject: " + self.email_subject

        if useMime:
            body = "This is a multi-part message in MIME format.\n"
            # Add text/html MIME part
            body = body + "--" + mime_boundary + "\n"
            body = body + "Content-Type: text/html; charset=UTF-8\n"
            body = body + "Content-Transfer-Encoding: quoted-printable\n\n"
            body = (
                body
                + quopri.encodestring(self.email_body.encode("utf-8")).decode("utf-8")
                + "\n"
            )

            # Add figures
            for fig in self.email_figlist:
                plt.figure(fig)
                (fd, fname) = tempfile.mkstemp(suffix=".png")
                f_png = os.fdopen(fd, "wb")
                plt.savefig(f_png)
                f_png.close()
                with open(fname, "rb") as image_file:
                    encoded_figure = base64.b64encode(image_file.read()).decode("ascii")
                os.remove(fname)
                # Add image/png MIME part
                body = body + "--" + mime_boundary + "\n"
                body = body + "Content-Type: image/png\n"
                body = body + "Content-Disposition: inline\n"
                body = body + "Content-Transfer-Encoding: base64\n\n"
                for i in range(0, len(encoded_figure), 76):
                    debut = i
                    fin = i + 75
                    if fin >= len(encoded_figure):
                        fin = len(encoded_figure) - 1
                    body = body + encoded_figure[debut : (fin + 1)] + "\n"

            # Send email
            try:
                error_list = smtp.sendmail(
                    self.email_from_addr,
                    self.email_to_addrs,
                    email_header + "\n" + body + "\n" + "--" + mime_boundary + "--\n",
                )
                if len(error_list) == 0:
                    success = True
            except smtplib.SMTPHeloError:
                print("SMTP Helo Error")
                pass
            except smtplib.SMTPRecipientsRefused:
                print("Some recipients have been rejected by SMTP server")
                pass
            except smtplib.SMTPSenderRefused:
                print("SMTP server refused sender " + self.email_from_addr)
                pass
            except smtplib.SMTPDataError:
                print("SMTP Data Error")
                pass

        else:
            try:
                email_content = (
                    email_header.encode("utf-8")
                    + b"\n"
                    + quopri.encodestring(self.email_body.encode("utf-8"))
                )
                error_list = smtp.sendmail(
                    self.email_from_addr, self.email_to_addrs, email_content
                )
                if len(error_list) == 0:
                    success = True
            except smtplib.SMTPHeloError:
                print("SMTP Helo Error")
                pass
            except smtplib.SMTPRecipientsRefused:
                print("Some recipients have been rejected by SMTP server")
                pass
            except smtplib.SMTPSenderRefused:
                print("SMTP server refused sender " + self.email_from_addr)
                pass
            except smtplib.SMTPDataError:
                print("SMTP Data Error")
                pass

        smtp.quit()
        self.email_body = ""
        self.email_started = False
        self.email_figlist = []
        if success:
            self.parameters["email_lastSent"] = time.time()
            date_string = time.strftime(
                dateformat, time.localtime(self.parameters["email_lastSent"])
            )
            print(date_string + ": Email successfully sent.")

    def time_since_last_email(self):
        try:
            last = self.parameters["email_lastSent"]
        except Exception:
            last = 0.0
            pass
        return time.time() - last

    def Stop(self):
        if self.email_started:
            self.stop_email()
            print("MI: email stopped.")
        if self.opened:
            self.store.close()
            self.datfile.close()
            date_string = time.strftime(dateformat, time.localtime(time.time()))
            self.logfile.write("Session closed on " + date_string)

            self.logfile.flush()
            self.logfile.close()
            self.opened = False
            print("MI: Successfully ended.")

    def __exit__(self, t, v, cb):
        self.Stop()
        super(Session, self).__exit__(t, v, cb)

    def __del__(self):
        if hasattr(self, "exited"):
            # let __enter__/__exit__ call self.Stop()
            pass
        else:
            # not inside a with statement, calling Stop when
            # object is being deleted
            self.Stop()



"""Photometrics module (:mod:`pymanip.video.photometrics`)
==========================================================

This module is a shortcut for the :class:`pymanip.video.photometrics.camera.Photometrics_Camera`
class.

"""

from .camera import Photometrics_Camera


"""
Module for legacy OctMI sessions

The OctSession class is designed to be used
in a similar fashion to SavedSession for
use by the collection classes.

"""

import os.path
import inspect
import h5py

from fluiddyn.util.terminal_colors import cprint
from .octmi_binary import read_OctMI_session


class OctSession(object):
    def __init__(
        self, session_name, cache_override=False, cache_location=".", verbose=False
    ):
        self.session_name = session_name
        self.variables = read_OctMI_session(session_name, verbose=verbose)
        self.time = self.variables.pop("t")
        self.cachestorename = (
            os.path.join(
                os.path.realpath(cache_location),
                "cache",
                os.path.basename(self.session_name),
            )
            + ".hdf5"
        )
        if cache_override:
            self.cachemode = "w"
        else:
            self.cachemode = "r+"
        try:
            self.cachestore = h5py.File(self.cachestorename, self.cachemode)
            if verbose:
                cprint.yellow("*** Cache store found at " + self.cachestorename)
            self.has_cachestore = True
        except IOError:
            self.has_cachestore = False
            pass

    def __str__(self):
        return self.session_name

    def describe(self):
        if len(self.variables) > 0:
            print("List of saved variables: (%d lines)" % len(self.time))
            for key, var in self.variables.items():
                print(" " + key)

    def has_dataset(self, name):
        return False

    def dataset(self, name):
        return None

    def has_parameter(self, name):
        return False

    def parameter(self, name):
        return None

    def has_log(self, name):
        if name in ["Time", "time", "t"]:
            return True
        return name in self.variables.keys()

    def log_variable_list(self):
        return self.variables.keys()

    def log(self, varname):
        if varname in ["Time", "time", "t"]:
            return self.time
        elif varname == "?":
            print(self.log_variable_list())
        elif varname in self.variables:
            return self.variables[varname]
        else:
            cprint.red("Variable is not defined: " + varname)

    def __getitem__(self, key):
        return self.log(key)

    @property
    def cachedvars(self):
        if not hasattr(self, "has_cachestore"):
            return []
        if self.has_cachestore:
            return self.cachestore.keys()
        else:
            return []

    def cached(self, *args):
        """
        cached('var1', 'var2', ...)
        or
        cached(['var1', 'var2'])

        returns True if all specified variables are present in the cachestore
        """
        if len(args) == 1:
            name = args[0]
        else:
            name = [a for a in args]
        if isinstance(name, str):
            return name in self.cachedvars
        else:
            return all([(n in self.cachedvars) for n in name])

    def cachedvalue(self, varname):
        if self.has_cachestore:
            cprint.yellow("Retriving " + varname + " from cache")
            return self.cachestore[varname].value
        else:
            return None

    def cache(self, name, dict_caller=None):
        if dict_caller is None:
            stack = inspect.stack()
            try:
                dict_caller = stack[1][0].f_locals
            finally:
                del stack
        if not isinstance(name, str):
            for var in name:
                self.cache(var, dict_caller)
            return
        if not self.has_cachestore:
            try:
                os.mkdir(os.path.dirname(self.cachestorename))
            except OSError:
                pass
            try:
                self.cachestore = h5py.File(self.cachestorename, "w")
                self.has_cachestore = True
                cprint.yellow("*** Cache store created at " + self.cachestorename)
            except IOError as ioe:
                self.has_cachestore = False
                cprint.red("Cannot create cache store")
                cprint.red(ioe.message)

        if self.has_cachestore:
            cprint.yellow("Saving " + name + " in cache")
            try:
                new_length = len(dict_caller[name])
            except TypeError:
                new_length = 1
                pass
            if name in self.cachestore.keys():
                if len(self.cachestore[name]) != new_length:
                    self.cachestore[name].resize((new_length,))
                self.cachestore[name][:] = dict_caller[name]
            else:
                if new_length > 1:
                    self.cachestore.create_dataset(
                        name, chunks=True, maxshape=(None,), data=dict_caller[name]
                    )
                else:
                    self.cachestore.create_dataset(
                        name, chunks=True, maxshape=(None,), shape=(new_length,)
                    )
                    self.cachestore[name][:] = dict_caller[name]

    def __del__(self):
        if hasattr(self, "has_cachestore") and self.has_cachestore:
            self.cachestore.close()



"""

API to query properties on a given resource.
- NISysCfgGetResourceProperty
- NISysCfgGetResourceIndexedProperty

"""


import ctypes
from pymanip.nisyscfg._lib import lib
from pymanip.nisyscfg._lib.constants import (
    NISysCfgResourceProperty,
    NISysCfgIndexedProperty,
    NISYSCFG_SIMPLE_STRING_LENGTH,
    NISysCfgStatus,
)
from pymanip.nisyscfg._lib.types import NISysCfgResourcePropertyType


def NISysCfgGetResourceProperty(resourceHandle, propertyID):
    (
        attr_ctype,
        attr_ini,
        create_func,
        ref_func,
        enum_class,
    ) = NISysCfgResourcePropertyType[propertyID]
    if create_func is None:
        create_func = attr_ctype
    if ref_func == ctypes.byref:
        attr_func = ctypes.POINTER(attr_ctype)
    else:
        attr_func = attr_ctype
    f = lib.NISysCfgGetResourceProperty
    f.argtypes = (ctypes.c_void_p, ctypes.c_int, attr_func)
    f.restype = ctypes.c_ulong
    val = create_func(attr_ini)
    status = f(resourceHandle, propertyID, ref_func(val))
    try:
        val = val.value
        val = val.decode("ascii")
    except AttributeError:
        pass

    return NISysCfgStatus(status), val


def NISysCfgGetResourceIndexedProperty(resourceHandle):
    f = lib.NISysCfgGetResourceIndexedProperty
    f.argtypes = (ctypes.c_void_p, ctypes.c_int, ctypes.c_uint, ctypes.c_char_p)
    f.restype = ctypes.c_ulong

    # implémenté seulement pour IndexedPropertyExpertName
    status, num_experts = NISysCfgGetResourceProperty(
        resourceHandle, NISysCfgResourceProperty.NumberOfExperts
    )
    if status != 0:
        raise RuntimeError(
            f"NISysCfgGetResourceProperty returned {str(NISysCfgStatus(status)):}"
        )
    data = list()
    for index in range(num_experts):
        indexed_data = dict()
        for attr in NISysCfgIndexedProperty:
            val = ctypes.create_string_buffer(NISYSCFG_SIMPLE_STRING_LENGTH)
            status = f(resourceHandle, attr, index, val)
            if status != 0:
                raise RuntimeError(
                    f"NISysCfgGetResourceIndexedProperty returned {str(NISysCfgStatus(status)):}"
                )
            indexed_data[attr.name] = val.value.decode("ascii")
        data.append(indexed_data)
    return data



"""Time utilities (:mod:`pymanip.mytime`)
=========================================

This modules contains very simple functions for handling dates and times.
In particular, the :func:`datestr2epoch` is useful to read fluidlab's session timestamps
which are roughly in RFC 3339 format.

.. autofunction:: sleep

.. autofunction:: datestr2epoch

.. autofunction:: epoch2datestr

.. autofunction:: tic

.. autofunction:: toc

"""

import datetime
from time import time
import sys
import warnings
from platform import platform
import dateutil.parser
from dateutil.tz import tzutc, tzlocal
import matplotlib.pyplot as plt
import six

if platform().startswith("Windows"):
    dateformat = "%A %d %B %Y - %X (%z)"
else:
    dateformat = "%A %e %B %Y - %H:%M:%S (UTC%z)"


def sleep(duration):
    """Prints a timer for specified duration.
    This is mostly similar to :meth:`pymanip.asyncsession.AsyncSession.sleep`, except that
    it is meant to be used outside a session.

    :param duration: the duration for which to sleep
    :type duration: float
    """
    debut = time()
    while (time() - debut) < duration:
        if six.PY2:
            sys.stdout.write(
                (
                    "Sleeping for "
                    + str(-int(time() - debut - duration))
                    + " s"
                    + " " * 35
                    + "\r"
                ).encode("utf-8")
            )
        else:
            sys.stdout.write(
                (
                    "Sleeping for "
                    + str(-int(time() - debut - duration))
                    + " s"
                    + " " * 35
                    + "\r"
                )
            )
        sys.stdout.flush()
        # time.sleep(1.0)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            plt.pause(1.0)
    sys.stdout.write("\n")


def datestr2epoch(string):
    """Convert datestr into epoch.
    Correct string is:
    '2016-02-25T17:36+0100' or '2016-02-25T17:36+01:00'.

    UTC+1 or UTC+0100 is also accepted by this function
    Accepts a single string or a list of string
    """
    if isinstance(string, str):
        string = string.upper()
        # Ci-dessous, on remplace UTC+1, UTC+01, UTC+0100 par +0100
        if "UTC" in string:
            parts = string.split("UTC")
            offset = parts[1].replace(":", "")
        if len(offset) == 0:
            offset = "+0000"
        elif int(offset) <= 12:
            offset = "{:+03d}00".format(int(offset))
        string = parts[0] + offset
        # print('New string: ', string)
        date = dateutil.parser.parse(string)
        if not date.tzinfo:
            raise TypeError("datestr with no timezone information " "is ambiguous !")
        return (date - datetime.datetime(1970, 1, 1, tzinfo=tzutc())).total_seconds()
    else:
        dlist = []
        for s in string:
            s = s.upper()
            # Ci-dessous, on remplace UTC+1, UTC+01, UTC+0100 par +0100
            if "UTC" in s:
                parts = s.split("UTC")
                offset = parts[1].replace(":", "")
            if len(offset) == 0:
                offset = "+0000"
            elif int(offset) <= 12:
                offset = "{:+03d}00".format(int(offset))
            s = parts[0] + offset
            # print('New string: ', string)
            date = dateutil.parser.parse(s)
            if not date.tzinfo:
                raise TypeError(
                    "datestr with no timezone information " "is ambiguous !"
                )
            dlist.append(
                (date - datetime.datetime(1970, 1, 1, tzinfo=tzutc())).total_seconds()
            )
        return dlist


def epoch2datestr(epoch, tz=None):
    """Convert epoch into datestr.
    If no tz is given, the output is given in Coordinated Universal Time
    """
    if not tz:
        # print("Swiching to UTC")
        tz = tzutc()
    date = datetime.datetime.fromtimestamp(epoch, tz=tz)
    return date.isoformat()


tic_starts = []


def tic():
    """Simple timer start.
    """
    tic_starts.append(time())


def toc(comment=None):
    """Simpler timer stop.

    :param comment: comment for print
    :type comment: str, optional
    """
    if comment is None:
        print("Elapsed time {:.2f} seconds.".format(time() - tic_starts[-1]))
    else:
        print(comment, "{:.2f} seconds.".format(time() - tic_starts[-1]))


if __name__ == "__main__":
    now = time()
    print("Now in local TZ:", epoch2datestr(now, tz=tzlocal()))
    print("Now in UTC:", epoch2datestr(now))
    reference = datestr2epoch("2016-02-25T17:36UTC")
    for datestr in [
        "2016-02-25T17:36UTC",
        "2016-02-25T18:36UTC+1",
        "2016-02-25T18:36UTC+01",
        "2016-02-25T18:36UTC+0100",
        "2016-02-25T19:36UTC+02:00",
        "2016-02-25T18:36+0100",
        "2016-02-25T17:36+0000",
    ]:
        epoch = datestr2epoch(datestr)
        datestr2 = epoch2datestr(epoch)
        epoch2 = datestr2epoch(datestr2)
        if epoch == epoch2 and epoch == reference:
            print("{:s} {:}".format(datestr, epoch))



"""Asynchronous acquisition session (:mod:`pymanip.video.session`)
==================================================================

This module provides a high-level class, :class:`VideoSession`, to be used in acquisition scripts.
Users should subclass :class:`VideoSession`, and add two methods: ``prepare_camera(self, cam)`` in which
additionnal camera setup may be done, and ``process_image(self, img)`` in which possible image post-processing
can be done.
The ``prepare_camera`` method, if defined, is called before starting camera acquisition.
The ``process_image`` method, if defined, is called before saving the image.

The implementation allows simultaneous acquisition of several cameras, trigged by a function generator.
Concurrent tasks are set up to grab images from the cameras to RAM memory, while a background task saves
the images to the disk. The cameras, and the trigger, are released as soon as acquisition is finished, even if the images are
still being written to the disk, which allows several scripts to be executed concurrently (if the host computer
has enough RAM). The ``trigger_gbf`` object must implement ``configure_square()`` and ``configure_burst()`` methods
to configure square waves and bursts, as well as ``trigger()`` method for software trigger. An exemple is
:class:`fluidlab.instruments.funcgen.agilent_33220a.Agilent33220a`.

A context manager must be used to ensure proper saving of the metadata to the database.

**Example**

.. code:: python

   import numpy as np
   import cv2

   from pymanip.video.session import VideoSession
   from pymanip.video.ximea import Ximea_Camera
   from pymanip.instruments import Agilent33220a


   class TLCSession(VideoSession):

       def prepare_camera(self, cam):
           # Set up camera (keep values as custom instance attributes)
           self.exposure_time = 50e-3
           self.decimation_factor = 2

           cam.set_exposure_time(self.exposure_time)
           cam.set_auto_white_balance(False)
           cam.set_limit_bandwidth(False)
           cam.set_vertical_skipping(self.decimation_factor)
           cam.set_roi(1298, 1833, 2961, 2304)

           # Save some metadata to the AsyncSession underlying database
           self.save_parameter(
               exposure_time=self.exposure_time,
               decimation_factor=self.decimation_factor,
           )

       def process_image(self, img):
           # On decimate manuellement dans la direction horizontale
           img = img[:, ::self.decimation_factor, :]

           # On redivise encore tout par 2
           # image_size = (img.shape[0]//2, img.shape[1]//2)
           # img = cv2.resize(img, image_size)

           # Correction (fixe) de la balance des blancs
           kR = 1.75
           kG = 1.0
           kB = 2.25
           b, g, r = cv2.split(img)
           img = np.array(
               cv2.merge([kB*b, kG*g, kR*r]),
               dtype=img.dtype,
           )

           # Rotation de 180°
           img = cv2.rotate(img, cv2.ROTATE_180)

           return img


   with TLCSession(
       Ximea_Camera(),
       trigger_gbf=Agilent33220a("USB0::0x0957::0x0407::SG43000299::INSTR"),
       framerate=24,
       nframes=1000,
       output_format="png",
       ) as sesn:

       # ROI helper (remove cam.set_roi in prepare_camera for correct usage)
       # sesn.roi_finder()

       # Single picture test
       # sesn.show_one_image()

       # Live preview
       # ts, count = sesn.live()

       # Run actual acquisition
       ts, count = sesn.run(additionnal_trig=1)

.. autoclass:: VideoSession
   :members:
   :private-members:
   :show-inheritance:

"""

import asyncio
from queue import SimpleQueue
from pathlib import Path
from datetime import datetime, timedelta

import numpy as np
import matplotlib.pyplot as plt
from progressbar import ProgressBar
import cv2

from pymanip.asyncsession import AsyncSession
from pymanip.video import MetadataArray


class VideoSession(AsyncSession):
    """This class represents a video acquisition session.

    :param camera_or_camera_list: Camera(s) to be acquired
    :type camera_or_camera_list: :class:`~pymanip.video.Camera` or list of :class:`~pymanip.video.Camera`
    :param trigger_gbf: function generator to be used as trigger
    :type trigger_gbf: :class:`~fluidlab.instruments.drivers.Driver`
    :param framerate: desired framerate
    :type framerate: float
    :param nframes: desired number of frames
    :type nframes: int
    :param output_format: desired output image format, "bmp", "png", tif", or video format "mp4"
    :type output_format: str
    :param output_format_params: additionnal params to be passed to :func:`cv2.imwrite`
    :type output_format_params: list
    :param exist_ok: allows to override existing output folder
    :type exist_ok: bool

    """

    def __init__(
        self,
        camera_or_camera_list,
        trigger_gbf,
        framerate,
        nframes,
        output_format,
        output_format_params=None,
        output_path=None,
        exist_ok=False,
        timeout=None,
        burst_mode=True,
    ):
        """Constructor method"""
        if isinstance(camera_or_camera_list, (list, tuple)):
            self.camera_list = camera_or_camera_list
        else:
            self.camera_list = (camera_or_camera_list,)
        self.trigger_gbf = trigger_gbf
        self.framerate = framerate
        self.nframes = nframes
        if output_format not in ("png", "tif", "mp4", "jpg", "bmp"):
            raise ValueError("output_format must be png, tif, jpg, bmp, or mp4")
        self.output_format = output_format
        self.output_format_params = output_format_params
        self.timeout = timeout
        self.burst_mode = burst_mode
        
        if output_path is None:
            current_date = datetime.today().strftime("%Y-%m-%d")
            base_dir = Path(".") / current_date
            base_dir.mkdir(exist_ok=True)

        if output_format == "mp4":
            if output_path is None:
                i = 1
                while True:
                    self.output_paths = [
                        base_dir / f"{i:02d}-cam{no:}.mp4"
                        for no in range(len(self.camera_list))
                    ]
                    session_name = base_dir / f"{i:02d}"
                    for ff in self.output_paths:
                        if ff.exists():
                            i = i + 1
                            break
                    else:
                        break
            else:
                self.output_paths = [
                    Path(output_path + f"-cam{no:}.mp4")
                    for no in range(len(self.camera_list))
                ]
                for ff in self.output_file:
                    if self.output_file.exists():
                        if exist_ok:
                            self.output_file.unlink()
                        else:
                            raise RuntimeError(
                                f"The file {self.output_file:} already exists."
                            )
                session_name = str(self.output_file)
                if session_name.endswith(".mp4"):
                    session_name = session_name[:-4]
            if len(self.output_paths) == 1:
                print("Output file is", self.output_paths[0])
            else:
                print("Output files are")
                for ff in self.output_paths:
                    print(str(ff))
        else:
            if output_path is None:
                i = 1
                while True:
                    self.output_folder = base_dir / f"{i:02d}"
                    session_name = self.output_folder / "session"
                    if self.output_folder.exists():
                        i = i + 1
                    else:
                        self.output_folder.mkdir()
                        break
            else:
                self.output_folder = Path(output_path)
                self.output_folder.mkdir(exist_ok=exist_ok)
                session_name = self.output_folder
            print("Output folder is", str(self.output_folder.absolute()))

        super().__init__(session_name, delay_save=True)
        for cam in self.camera_list:
            if self.trigger_gbf is not None or self.trigger_gbf is True:
                cam.set_trigger_mode(True)
            else:
                cam.set_trigger_mode(False)
                cam.set_frame_rate(framerate)
        if self.trigger_gbf is True:
            self.trigger_gbf = None
        self.image_queues = [SimpleQueue() for _ in self.camera_list]
        self.acquisition_finished = [False] * len(self.camera_list)
        self.initialising_cams = set(self.camera_list)

    async def _acquire_images(self, cam_no):
        """Private instance method: image acquisition task.
        This task asynchronously iterates over the given camera frames, and puts the obtained images
        in a simple FIFO queue.

        :param cam_no: camera index
        :type cam_no: int
        """
        with self.camera_list[cam_no] as cam:
            if hasattr(self, "prepare_camera"):
                self.prepare_camera(cam)
            if cam_no == 0 and np.isfinite(self.nframes):
                pb = ProgressBar(initial_value=0, min_value=0, max_value=self.nframes)
            else:
                pb = None
            gen = cam.acquisition_async(
                self.nframes,
                initialising_cams=self.initialising_cams,
                raise_on_timeout=False,
                timeout=self.timeout,
            )
            n = 0
            async for im in gen:
                if im is not None:
                    self.image_queues[cam_no].put(im.copy())
                    n = n + 1
                    if pb is not None:
                        pb.update(n)
                else:
                    print("Camera timed out! Stopping...")
                    self.running = False
                if not self.running:
                    success = await gen.asend(True)
                    if not success:
                        print("Unable to stop camera acquisition")
            if pb is not None:
                pb.finish()
            self.acquisition_finished[cam_no] = True
            if all(self.acquisition_finished):
                self.running = False
        print(f"{n:d} images acquired (cam {cam_no:}).")

    async def _start_clock(self):
        """Private instance method: clock starting task.
        This task waits for all the cameras to be ready for trigger, and then sends a software trig to
        the function generator.
        """
        print("clock task")
        while len(self.initialising_cams) > 0:
            await asyncio.sleep(0.1)
        with self.trigger_gbf:
            self.trigger_gbf.trigger()
            print("trigger sent")
        return datetime.now().timestamp()

    async def _save_images(self, keep_in_RAM=False, unprocessed=False, no_save=False):
        """Private instance method: image saving task.
        This task checks the image FIFO queue. If an image is available, it is taken out of the queue and
        saved to the disk.

        :param keep_in_RAM: if set, images are not saved and kept in a list.
        :type keep_in_RAM: bool
        :param unprocessed: if set, the ``process_image`` method is not called.
        :type unprocessed: bool
        :param no_save: do not actually save (dry run), for testing purposes
        :type no_save: bool
        """
        loop = asyncio.get_event_loop()
        i = [0] * len(self.camera_list)
        pb = None

        if keep_in_RAM:
            self.image_list = [list() for _ in range(len(self.camera_list))]

        while True:
            if not self.running:
                if all([q.empty() for q in self.image_queues]):
                    break
            for cam_no, q in enumerate(self.image_queues):
                if q.empty():
                    await asyncio.sleep(0.1)
                else:
                    img = q.get()
                    self.add_entry(
                        ts=img.metadata["timestamp"], count=img.metadata["counter"]
                    )
                    if not no_save:
                        if hasattr(self, "process_image") and not unprocessed:
                            img = await loop.run_in_executor(None, self.process_image, img)
                        filepath = (
                            self.output_folder
                            / f"img-cam{cam_no:d}-{(i[cam_no]+1):04d}.{self.output_format:}"
                        )
                    if keep_in_RAM:
                        self.image_list[cam_no].append(img)
                    elif not no_save:
                        await loop.run_in_executor(
                            None,
                            cv2.imwrite,
                            str(filepath),
                            img,
                            self.output_format_params,
                        )
                    i[cam_no] += 1
            if not self.running:
                i_min = min(i)
                if pb is None:
                    print("Saving is continuing...")
                    pb = ProgressBar(
                        min_value=0, max_value=self.nframes, initial_value=i_min
                    )
                else:
                    pb.update(i_min)
        if pb is not None:
            pb.finish()
        for cam_no, ii in enumerate(i):
            print(f"{ii:d} images saved (cam {cam_no:}).")

    def _convert_for_ffmpeg(self, cam_no, img, fmin, fmax, gain):
        """Private instance method: image conversion for ffmpeg process.
        This method prepares the input image to bytes to be sent to the ffmpeg pipe.

        :param cam_no: camera index
        :type cam_no: int
        :param img: image to process
        :type img: :class:`numpy.ndarray`
        :param fmin: minimum level
        :type fmin: int
        :param fmax: maximum level
        :type fmax: int
        :param gain: gain
        :type gain: float
        """
        fff = 255 * gain * np.array(img - fmin, dtype=np.float64) / (fmax - fmin)
        fff[fff < 0] = 0
        fff[fff > 255] = 255
        ff = np.array(fff, dtype=np.uint8)
        if self.camera_list[cam_no].color_order == "RGB":
            ff = cv2.cvtColor(ff, cv2.COLOR_RGB2BGR)
        return ff.tostring()

    async def _save_video(self, cam_no, gain=1.0, unprocessed=False):
        """Private instance method: video saving task.
        This task waits for images in the FIFO queue, and sends them to ffmpeg via a pipe.

        :param cam_no: camera index
        :type cam_no: int
        :param gain: gain
        :type gain: float
        :param unprocessed: if set, :meth:`process_image` method is not called
        :type unprocessed: bool
        """
        loop = asyncio.get_event_loop()
        command = None
        fmin = None
        fmax = None
        while self.running or not self.image_queues[cam_no].empty():
            if self.image_queues[cam_no].empty():
                await asyncio.sleep(0.1)
            else:
                img = self.image_queues[cam_no].get()
                self.add_entry(
                    ts=img.metadata["timestamp"], count=img.metadata["counter"]
                )
                if hasattr(self, "process_image") and not unprocessed:
                    img = await loop.run_in_executor(None, self.process_image, img)
                if command is None:
                    output_size = img.shape
                    command = [
                        "-y",
                        "-f",
                        "rawvideo",
                        "-vcodec",
                        "rawvideo",
                        "-s",
                        "{}x{}".format(output_size[1], output_size[0]),
                        "-pix_fmt",
                        "bgr24",
                        "-r",
                        str(self.framerate),
                        "-i",
                        "-",
                        "-an",
                        "-vcodec",
                        "mpeg4",
                        "-b:v",
                        "5000k",
                        str(self.output_paths[cam_no]),
                    ]
                    proc = await asyncio.create_subprocess_exec(
                        "ffmpeg",
                        *command,
                        stdin=asyncio.subprocess.PIPE,
                        stdout=None,
                        stderr=None,
                    )
                if fmin is None:
                    fmin = np.min(img)
                if fmax is None:
                    fmax = np.max(img)
                data = await loop.run_in_executor(
                    None, self._convert_for_ffmpeg, cam_no, img, fmin, fmax, gain
                )
                proc.stdin.write(data)
                await proc.stdin.drain()
        proc.stdin.close()
        await asyncio.wait_for(proc.stdin.wait_closed(), timeout=5.0)
        await asyncio.wait_for(proc.wait(), timeout=5.0)
        print("ffmpeg has terminated.")

    async def _fast_acquisition_to_ram(self, cam_no, total_timeout_s):
        """Private instance method: fast acquisition to ram task"""
        with self.camera_list[cam_no] as cam:
            if hasattr(self, "prepare_camera"):
                self.prepare_camera(cam)
            ts, count, images = await cam.fast_acquisition_to_ram(
                self.nframes,
                total_timeout_s,
                self.initialising_cams,
                False,  # raise_on_timeout
            )
        return ts, count, images

    async def main(
        self,
        keep_in_RAM=False,
        additionnal_trig=0,
        live=False,
        unprocessed=False,
        delay_save=False,
        no_save=False,
    ):
        """Main entry point for acquisition tasks. This asynchronous task can be called
        with :func:`asyncio.run`, or combined with other user-defined tasks.

        :param keep_in_RAM: do not save to disk, but keep images in a list
        :type keep_in_RAM: bool
        :param additionnal_trig: additionnal number of pulses sent to the camera
        :type additionnal_trig: int
        :param live: toggle live preview
        :type live: bool
        :param unprocessed: do not call :meth:`process_image` method.
        :type unprocessed: bool
        :return: camera_timestamps, camera_counter
        :rtype: :class:`numpy.ndarray`, :class:`numpy.ndarray`
        """
        if self.trigger_gbf is not None:
            with self.trigger_gbf:
                if live or self.nframes < 2 or not self.burst_mode:
                    self.trigger_gbf.configure_square(0.0, 5.0, freq=self.framerate)
                else:
                    self.trigger_gbf.configure_burst(
                        self.framerate, self.nframes + additionnal_trig
                    )
                    self.save_parameter(fps=self.framerate, num_imgs=self.nframes)

        if delay_save and all(
            [
                hasattr(self.camera_list[cam_no], "fast_acquisition_to_ram")
                for cam_no in range(len(self.camera_list))
            ]
        ):
            if live:
                raise ValueError("delay_save and live are not compatible")
            print("Acquisition in fast mode to RAM")

            # Acquisition to RAM
            total_timeout_s = 5 + (self.nframes / self.framerate)
            acquisition_tasks = [
                self._fast_acquisition_to_ram(cam_no, total_timeout_s)
                for cam_no in range(len(self.camera_list))
            ]
            dt_start = datetime.now()
            dt_end = dt_start + timedelta(seconds=total_timeout_s)
            print("Acquisition start time:", dt_start.strftime("%d-%m-%Y %H:%M:%S"))
            print("              end time:", dt_end.strftime("%d-%m-%Y %H:%M:%S"))
            results = await asyncio.gather(*acquisition_tasks, self._start_clock())
            self.running = False

            # Convert from lists to queues (no data copy involved)
            for cam_no, (ts_all, count_all, images_all) in zip(
                range(len(self.camera_list)), results
            ):
                for ts, count, image in zip(ts_all, count_all, images_all):
                    self.image_queues[cam_no].put(
                        MetadataArray(
                            image,
                            metadata={
                                "counter": count,
                                "timestamp": ts,
                            },
                        )
                    )
        else:
            acquisition_tasks = [
                self._acquire_images(cam_no) for cam_no in range(len(self.camera_list))
            ]
            if live:
                save_tasks = [self._live_preview(unprocessed)]
            elif self.output_format == "mp4":
                save_tasks = [self._start_clock()]
                if not delay_save:
                    save_tasks = save_tasks + [
                        self._save_video(cam_no, unprocessed=unprocessed)
                        for cam_no in range(len(self.camera_list))
                    ]

            else:
                if self.trigger_gbf is not None:
                    save_tasks = [self._start_clock()]
                else:
                    save_tasks = []
                if not delay_save:
                    save_tasks = save_tasks + [
                        self._save_images(keep_in_RAM, unprocessed)
                    ]

            await self.monitor(*acquisition_tasks, *save_tasks, server_port=None)

        # Post-acquisition save if the save_tasks were not included (in fast mode, or in regular mode)
        if delay_save:
            if self.output_format == "mp4":
                for cam_no in range(len(self.camera_list)):
                    await self._save_video(cam_no, unprocessed=unprocessed)
            else:
                await self._save_images(keep_in_RAM, unprocessed, no_save)

        # Post-acquisition information
        _, camera_timestamps = self.logged_variable("ts")
        _, camera_counter = self.logged_variable("count")
        
        if isinstance(camera_timestamps[0], str):
            camera_timestamps = np.array([datetime.fromisoformat(ts).timestamp() for ts in camera_timestamps])

        if camera_timestamps.size > 2:
            dt = camera_timestamps[1:] - camera_timestamps[:-1]
            mean_dt = np.mean(dt)
            mean_fps = 1.0 / mean_dt
            min_dt = np.min(dt)
            max_dt = np.max(dt)
            min_fps = 1.0 / max_dt
            max_fps = 1.0 / min_dt
            print(f"fps = {mean_fps:.3f} (between {min_fps:.3f} and {max_fps:.3f})")

        return camera_timestamps, camera_counter

    def run(self, additionnal_trig=0, delay_save=False, no_save=False):
        """Run the acquisition.

        :param additionnal_trig: additionnal number of pulses sent to the camera
        :type additionnal_trig: int
        :return: camera_timestamps, camera_counter
        :rtype: :class:`numpy.ndarray`, :class:`numpy.ndarray`
        """
        ts, count = asyncio.run(
            self.main(additionnal_trig=additionnal_trig, delay_save=delay_save, no_save=no_save)
        )
        return ts, count

    def live(self):
        """Starts live preview."""
        old_nframes = self.nframes
        old_output_format = self.output_format
        try:
            self.nframes = float("inf")
            self.output_format = "png"
            ts, count = asyncio.run(self.main(live=True))
        finally:
            self.nframes = old_nframes
            self.output_format = old_output_format
        return ts, count

    def get_one_image(
        self, additionnal_trig=0, unprocessed=False, unpack_solo_cam=True
    ):
        """Get one image from the camera(s).

        :param additionnal_trig: additionnal number of pulses sent to the camera
        :type additionnal_trig: int
        :param unprocessed: do not call :meth:`process_image` method.
        :type unprocessed: bool
        :param unpack_solo_cam: if set, keep list on return, even if there is only one camera
        :type unpack_solo_cam: bool
        :return: image(s) from the camera(s)
        :rtype: :class:`numpy.ndarray` or list of :class:`numpy.ndarray` (if multiple cameras, or ``unpack_solo_cam=False``).
        """
        old_nframes = self.nframes
        old_output_format = self.output_format
        try:
            self.nframes = 1
            self.output_format = "png"
            ts, count = asyncio.run(
                self.main(
                    keep_in_RAM=True,
                    additionnal_trig=additionnal_trig,
                    unprocessed=unprocessed,
                )
            )
            if len(self.camera_list) > 1 or not unpack_solo_cam:
                result = [im[0] for im in self.image_list]
            else:
                result = self.image_list[0][0]
        finally:
            self.nframes = old_nframes
            self.output_format = old_output_format
        return result

    def show_one_image(self, additionnal_trig=0):
        """Get one image from the camera(s), and plot them with matplotlib.

        :param additionnal_trig: additionnal number of pulses sent to the camera
        :type additionnal_trig: int
        """
        image = self.get_one_image(additionnal_trig)
        if isinstance(image, list):
            for img in image:
                plt.figure()
                if img.ndim == 2:
                    plt.imshow(img)
                    plt.colorbar()
                else:
                    plt.imshow(cv2.cvtColor(img, cv2.COLOR_BGR2RGB))
        else:
            plt.figure()
            if image.ndim == 2:
                plt.imshow(image)
                plt.colorbar()
            else:
                plt.imshow(cv2.cvtColor(image, cv2.COLOR_BGR2RGB))
        plt.show()

    def roi_finder(self, additionnal_trig=0):
        """Helper to determine the ROI. This method grabs one unprocessed image from the camera(s),
        and allows interactive selection of the region of interest.
        Attention: it is assumed that the :meth:`prepare_camera` method did not already set the ROI.

        :param additionnal_trig: additionnal number of pulses sent to the camera
        :type additionnal_trig: int
        :return: region of interest coordinates, X0, Y0, X1, Y1
        :rtype: list of floats
        """
        images = self.get_one_image(
            additionnal_trig, unprocessed=True, unpack_solo_cam=False
        )
        for cam_no, img in enumerate(images):
            try:
                l, c = img.shape
                color = False
            except ValueError:
                l, c, ncomp = img.shape
                color = True

            # Convert color order if necessary
            if color and self.camera_list[cam_no].color_order == "BGR":
                img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)

            plt.figure()
            plt.imshow(img)
            plt.title("Unprocessed image")
            print("Click on top-left and bottom-right corners")
            bottom_left, top_right = plt.ginput(2)
            X0, Y0 = bottom_left
            X1, Y1 = top_right
            print("ROI is", X0, Y0, X1, Y1)
        return X0, Y0, X1, Y1

    async def _live_preview(self, unprocessed=False):
        """Private instance method: live preview task.
        This task checks the FIFO queue, drains it and shows the last frame of each camera using cv2.
        If more than one frame is in the queues, the older frames are dropped.

        :param unprocessed: do not call :meth:`process_image` method.
        :type unprocessed: bool
        """
        minimum = None
        maximum = None
        while self.running or any([not q.empty() for q in self.image_queues]):
            for cam_no, q in enumerate(self.image_queues):
                ndropped = -1
                img = None
                while not q.empty():
                    img = q.get()
                    self.add_entry(
                        ts=img.metadata["timestamp"], count=img.metadata["counter"]
                    )
                    ndropped += 1
                if img is not None:
                    if ndropped > 0:
                        print(ndropped, "frames dropped.", end="\r")
                    if hasattr(self, "process_image") and not unprocessed:
                        img = self.process_image(img)

                    # Show only a 800x600 window (max)
                    try:
                        l, c = img.shape
                        color = False
                        minimum = np.min(img)
                        maximum = np.max(img)
                        maxint = np.iinfo(img.dtype).max
                    except ValueError:
                        l, c, ncomp = img.shape
                        color = True
                    zoom_l = l / 600
                    zoom_c = c / 800
                    zoom = max([zoom_l, zoom_c])
                    if zoom > 1:
                        img = cv2.resize(img, (int(c / zoom), int(l / zoom)))

                    # Convert color order if necessary
                    if color:
                        if self.camera_list[cam_no].color_order == "RGB":
                            img = cv2.cvtColor(img, cv2.COLOR_RGB2BGR)
                    else:
                        # if not in color, we rescale min max to first image (otherwise may appear black)
                        img = (maxint // (maximum - minimum)) * (img - minimum)

                    # Show image and run cv2 event loop
                    cv2.imshow(f"cam{cam_no:}", img)
            k = cv2.waitKey(1)
            if k != -1:
                self.ask_exit()
            await asyncio.sleep(0.1)
        cv2.destroyAllWindows()



import tkinter

try:
    from pymanip.aiodaq.daqmx import get_device_list as daqmx_get_devices

    has_daqmx = True
except ImportError:
    has_daqmx = False
try:
    from pymanip.aiodaq.scope import get_device_list as scope_get_devices

    has_scope = True
except ImportError:
    has_scope = False


class ChannelSelector:
    def __init__(self):
        if has_daqmx:
            self.daqmx_device_list = daqmx_get_devices()
        else:
            self.daqmx_device_list = dict()
        self.device_list = self.daqmx_device_list.copy()
        if has_scope:
            self.scope_device_list = scope_get_devices(self.daqmx_device_list)
            self.device_list.update(self.scope_device_list)
        else:
            self.scope_device_list = dict()
        self.channel_backend = dict()
        for name, devlist in self.daqmx_device_list.items():
            for dev in devlist:
                self.channel_backend[dev] = "daqmx"
        for name, devlist in self.scope_device_list.items():
            for dev in devlist:
                self.channel_backend[dev] = "scope"

    def print_channel_list(self):
        for name, devlist in self.device_list.items():
            print(name)
            print("-" * len(name))
            print(devlist)

    def gui_select(self):
        master = tkinter.Tk()
        ii = 0
        tkinter.Label(master, text="Choose channels").grid(row=ii, sticky=tkinter.W)
        ii += 1
        values = dict()
        for card, devlist in self.device_list.items():
            for dev in devlist:
                values[dev] = tkinter.IntVar()
                tkinter.Checkbutton(master, text=dev, variable=values[dev]).grid(
                    row=ii, sticky=tkinter.W
                )
                ii += 1
        tkinter.Button(master, text="OK", command=master.quit).grid(
            row=ii, sticky=tkinter.W
        )
        tkinter.mainloop()
        master.destroy()
        selected_channels = [chan for chan in values if values[chan].get()]
        selected_backends = [self.channel_backend[chan] for chan in selected_channels]
        if not all([b == selected_backends[0] for b in selected_backends]):
            raise ValueError("Channels from different backends were selected !")
        return selected_backends[0], selected_channels



"""Photometrics Camera module (:mod:`pymanip.video.photometrics.camera`)
========================================================================

This module implements the :class:`pymanip.video.photometrics.Photometrics_Camera` using the
`Python wrapper provided by Photometrics <https://github.com/Photometrics/PyVCAM>`_.
The documentation for the PVCAM SDK is available `online <https://www.photometrics.com/docs/pvcam-sdk/>`_.

.. autoclass:: Photometrics_Camera
   :members:
   :private-members:
   :show-inheritance:

"""

from time import monotonic
import asyncio
import numpy as np
from pymanip.video import MetadataArray, Camera, CameraTimeout

from pyvcam import pvc
from pyvcam.camera import Camera as PVCamera

pvc_initialized = False


class Photometrics_Camera(Camera):
    """Concrete :class:`pymanip.video.Camera` class for Photometrics camera."""

    def __init__(self, cam_num=0, readout_port=0):
        """Constructor method.

        readout_port is camera mode:
            - readout_port = 0 ("Sensitivity") 12 bits, max fps = 88
            - readout_port = 1 ("Speed") 8 bits, max fps = 500
            - readout_port = 2 ("Dynamic Range") 16 bits, max fps = 83
        """
        global pvc_initialized

        if not pvc_initialized:
            print("pvc.init_pvcam")
            pvc.init_pvcam()
            pvc_initialized = True

        for num, cam in enumerate(PVCamera.detect_camera()):
            if num == cam_num:
                break
        else:
            raise (f"Unable to find camera num {cam_num:}.")
        self.cam = cam
        self.cam.open()
        self.cam.readout_port = readout_port

        self.cam.meta_data_enabled = True
        print(self.cam.name)

    def close(self):
        """Close connection to the camera"""
        self.cam.close()
        self.cam = None

    def __exit__(self, type_, value, cb):
        """Context manager exit method"""
        super().__exit__(type_, value, cb)
        self.close()

    @property
    def name(self):
        return self.cam.name

    def set_trigger_mode(self, external):
        """Set external trigger (edge rising).
        Possible modes are available in self.cam.exp_modes.
        """
        if external:
            if self.cam.readout_port != 1:
                self.cam.exp_mode = "Edge Trigger"
            else:
                print("Fast mode cannot work with external trigger")
        else:
            self.cam.exp_mode = "Internal Trigger"

    def get_trigger_mode(self):
        if self.cam.exp_mode == "Internal Trigger":
            return False
        return True

    def set_exposure_time(self, seconds):
        """This method sets the exposure time for the camera.

        :param seconds: exposure in seconds.
        :type seconds: float
        """
        self.cam.exp_time = int(seconds * 1e3)
        print("Exposure time set to", 1000 * seconds, "ms")

    def get_exposure_time(self):
        """This method gets the exposure time in seconds for the camera."""
        return self.cam.exp_time / 1e3

    def set_roi(self, roiX0=0, roiY0=0, roiX1=0, roiY1=0):
        """This method sets the positions of the upper left corner (X0,Y0) and lower right
        (X1,Y1) corner of the ROI (region of interest) in pixels.
        """
        self.cam.roi = (roiX0, roiY0, roiX1, roiY1)

    async def get_image(self, loop, timeout):
        start_time = monotonic()
        while monotonic() - start_time < timeout:
            frameStatus = self.cam.check_frame_status()
            if frameStatus == "READOUT_FAILED":
                raise RuntimeError("Readout failed")
            else:
                future = loop.run_in_executor(
                    None,
                    pvc.get_frame,
                    self.cam._Camera__handle,
                    self.cam._Camera__shape[0],
                    self.cam._Camera__shape[1],
                    self.cam._Camera__bits_per_pixel,
                )
                try:
                    frame, fps, frame_count = await asyncio.wait_for(
                        future, timeout=timeout, loop=loop
                    )
                except asyncio.TimeoutError:
                    print("Timeout while waiting for poll_frame")
                    break
                frame["pixel_data"] = frame["pixel_data"].reshape(
                    self.cam._Camera__shape[1], self.cam._Camera__shape[0]
                )
                return (
                    frame,
                    fps,
                    frame_count,
                )

        print(
            "Out of the loop after",
            monotonic() - start_time,
            "second with frame status",
            frameStatus,
        )
        raise CameraTimeout

    async def acquisition_async(
        self,
        num=np.inf,
        timeout=None,
        raw=False,
        initialising_cams=None,
        raise_on_timeout=True,
    ):
        """Concrete implementation of :meth:`pymanip.video.Camera.acquisition_async` for the Photometrics camera.

        timeout in milliseconds.
        """
        loop = asyncio.get_event_loop()
        if timeout is None:
            timeout = max((5000, 5 * self.get_exposure_time() * 1000))
        try:
            count = 0
            """
            if np.isfinite(num):
                self.cam.start_seq(num_frames=num)
            else:
                self.cam.start_live()
            """
            self.cam.start_live()
            while count < num:
                if (
                    count == 0
                    and initialising_cams is not None
                    and self in initialising_cams
                ):
                    initialising_cams.remove(self)
                try:
                    frame, fps, frame_count = await self.get_image(loop, timeout / 1000)
                except CameraTimeout:
                    print("Camera timeout")
                    if raise_on_timeout:
                        raise
                    else:
                        stop_signal = yield None
                        if stop_signal:
                            break
                        else:
                            continue
                # d'après la doc, timestampBOF*timestampResN est en nanoseconds
                # mais il n'y a pas timestampResN, et il semble que c'est plutôt en picoseconds ?
                stop_signal = yield MetadataArray(
                    frame["pixel_data"],  # no copy
                    metadata={
                        "counter": frame_count,
                        "timestamp": frame["meta_data"]["frame_header"]["timestampBOF"]
                        / 1e12,
                    },
                )
                if count == 0:
                    for k, v in frame["meta_data"]["frame_header"].items():
                        print(k, v)
                    print("roi_headers", frame["meta_data"]["roi_headers"])
                count += 1
                if stop_signal:
                    break
        finally:
            self.cam.finish()
        if stop_signal:
            yield True

    async def empty_buffer(self, loop):
        start = monotonic()
        empty = False
        n = 0
        while (monotonic() - start) < 1.0:
            future = loop.run_in_executor(
                None,
                self.cam.poll_frame
            )
            try:
                _, _, _ = await asyncio.wait_for(future, timeout=0.5, loop=loop)
                n = n + 1
            except asyncio.TimeoutError:
                empty = True
                break
        print(n, "frames removed from circular buffer")
        if empty:
            print("Circular buffer successfully emptied")
        return empty

    async def fast_acquisition_to_ram(
        self, num, total_timeout_s=5*60, initialising_cams=None, raise_on_timeout=True
    ):
        """Fast method (without the overhead of run_in_executor and asynchronous generator), for acquisitions
        where concurrent saving is not an option (because the framerate is so much faster than writting time),
        so all frames are saved in RAM anyway.
        """

        count = np.empty((num,))
        ts = np.empty((num,))
        images = list()
        loop = asyncio.get_event_loop()
        try:
            self.cam.start_live()
            
            if initialising_cams and self.get_trigger_mode():
                await self.empty_buffer(loop)

            def _sync_loop():
                n = 0
                while n < num:
                    if (
                        n == 0
                        and initialising_cams is not None
                        and self in initialising_cams
                    ):
                        initialising_cams.remove(self)
                    frame, fps, frame_count = self.cam.poll_frame()
                    ts[n] = frame["meta_data"]["frame_header"]["timestampEOF"] / 1e12
                    count[n] = frame_count
                    n = n + 1
                    images.append(frame["pixel_data"])
            future = loop.run_in_executor(None, _sync_loop)
            try:
                await asyncio.wait_for(future, timeout=total_timeout_s, loop=loop)
            except asyncio.TimeoutError:
                print("Camera timeout")
                if raise_on_timeout:
                    raise CameraTimeout
            n = len(images)
            print(n, "frames read.")
        finally:
            self.cam.finish()

        return ts[:n], count[:n], images


if __name__ == "__main__":
    with Photometrics_Camera() as cam:
        print("Exposure =", cam.get_exposure_time())



#! /usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function

from pymanip import SavedSession
from pymanip.legacy_session import OctSession
import os
import six


class Manip(object):
    def __init__(
        self, session_name, nickname=None, directory=None, verbose=True, **kwargs
    ):
        self.properties = dict()
        for key in kwargs:
            self.properties[key] = kwargs[key]
        self.session_name = session_name
        if nickname:
            self.nickname = nickname
        else:
            self.nickname = session_name
        self.directory = directory
        self.verbose = verbose
        self.cachedvars = self.MI.cachedvars
        self.cached = self.MI.cached
        self.cachedvalue = self.MI.cachedvalue
        self.cache = self.MI.cache

    def __str__(self):
        return self.nickname

    def get(self, name):
        """
        Lookup for parameter of given name, first within the Manip object
        properties, and then inside the saved parameters of the specified
        session.
        """

        value = None
        if name == "basename" or name == "session_name":
            value = self.session_name
        elif name == "nickname":
            value = self.nickname
        elif name in self.properties:
            value = self.properties[name]
        else:
            if self.MI.has_log(name):
                value = self.MI.log(name)
            elif self.MI.has_dataset(name):
                value = self.MI.dataset(name)
            elif self.MI.has_parameter(name):
                value = self.MI.parameter(name)
            else:
                raise KeyError("Key not found {:s}".format(name))

        return value

    def dataset(self, name):
        return self.MI.dataset(name)

    def __getitem__(self, key):
        return self.get(key)

    def describe(self):
        self.MI.describe()

    @property
    def MI(self):
        if not hasattr(self, "_MI"):
            if self.directory:
                name = os.path.join(self.directory, self.session_name)
            else:
                name = self.session_name
            try:
                self._MI = SavedSession(name, verbose=self.verbose)
            except IOError as a:
                try:
                    self._MI = OctSession(name, verbose=self.verbose)
                except IOError as b:
                    print("None of the possible files can be found:")
                    print(" 1. {:}".format(a.filename))
                    print(" 2. {:}".format(b.filename))
                    raise IOError('Cannot open session "{:}" for reading'.format(name))
                # print('OctMI legacy mode')

        return self._MI


class ManipCollection(Manip):
    def __init__(self, basename, nickname=None, **kwargs):
        self.basename = basename
        self.num = kwargs.pop("num", 1)
        self.verbose = kwargs.pop("verbose", True)

        super(ManipCollection, self).__init__(
            session_name=basename, nickname=nickname, verbose=self.verbose, **kwargs
        )

    def __getitem__(self, key):
        if isinstance(key, six.string_types):
            return super(ManipCollection, self).__getitem__(key)
        else:
            try:
                if self.directory:
                    name = os.path.join(self.directory, self.basename + "_" + str(key))
                else:
                    name = self.basename + "_" + str(key)
                MI = SavedSession(name, verbose=self.verbose)
            except IOError as e:
                print('Unable to read file "' + str(e.filename) + "'.")
                print("Errno = " + str(e.errno))
                print("Message: " + str(e.message))
                raise IndexError
            return MI

    @property
    def MI(self):
        if hasattr(self, "current_acq"):
            return self[self.current_acq]
        return self[1]

    def items_from(self, start):
        # __iter__ est appelé dans tous les cas
        self.custom_start = start
        return self

    def __iter__(self):
        if hasattr(self, "custom_start"):
            self.current_acq = self.custom_start
            delattr(self, "custom_start")
        else:
            self.current_acq = 1
        return self

    def __next__(self):
        c = self.current_acq
        self.current_acq = c + 1
        if c > self.num:
            raise StopIteration
        else:
            return self.__getitem__(c)


class ManipList(object):
    def __init__(self, *args):
        if len(args) == 1:
            if isinstance(args[0], list):
                self.manips = args[0]
            else:
                self.manips = [args[0]]
        else:
            self.manips = args

    def __len__(self):
        return len(self.manips)

    def __iter__(self):
        return self.manips.__iter__()

    def __getitem__(self, key):
        if isinstance(key, six.string_types):
            return self.from_nickname(key)
        else:
            return self.manips.__getitem__(key)

    def from_nickname(self, nickname):
        return self.lookup(nickname=nickname)[0]

    def lookup(self, **kwargs):
        result = list()
        for m in self.manips:
            keep = True
            for key in kwargs:
                if m.get(key) != kwargs[key]:
                    keep = False
            if keep:
                result.append(m)
        return result



from ._lib.constants import NISysCfgLocale, NISysCfgStatus
from ._lib.session import NISysCfgInitializeSession, NISysCfgCloseHandle


class NISysCfgSession:
    def __init__(
        self,
        target="localhost",
        user=None,
        password=None,
        lang=NISysCfgLocale.Default,
        force_refresh=False,
        connect_timeout=1000,
    ):
        self.target = target
        self.user = user
        self.password = password
        self.lang = lang
        self.force_refresh = force_refresh
        self.connect_timeout = connect_timeout
        self.expertHandle = None
        self.sessionHandle = None

    def __enter__(self):
        status, expertHandle, sessionHandle = NISysCfgInitializeSession()
        if status != NISysCfgStatus.OK:
            raise RuntimeError(
                f"NISysCfgInitializeSession failed. Status: {str(status):}."
            )
        self.expertHandle = expertHandle
        self.sessionHandle = sessionHandle
        return self

    def __exit__(self, type_, value, cb):
        if self.expertHandle:
            status = NISysCfgCloseHandle(self.expertHandle)
            if status != NISysCfgStatus.OK:
                raise RuntimeError(
                    f"NISysCfgCloseHandle failed. Status: {str(status):}."
                )
        if self.sessionHandle:
            status = NISysCfgCloseHandle(self.sessionHandle)
            if status != NISysCfgStatus.OK:
                raise RuntimeError(
                    f"NISysCfgCloseHandle failed. Status: {str(status):}."
                )



"""Instruments module (:mod:`pymanip.instruments`)
==================================================

This module auto-imports all the instruments classes from :class:`fluidlab.instruments`:

.. currentmodule:: pymanip.instruments

.. autosummary::

{instruments:}

"""

from .fluidlab_instruments import *
from .fluidlab_instruments import __all__

instruments_str = "\n".join("   " + c for c in __all__)
__doc__ = __doc__.format(instruments=instruments_str)



"""Asynchronous extension of :class:`~fluidlab.interfaces.serial_inter.SerialInterface` (:mod:`pymanip.interfaces.aioserial`)
=============================================================================================================================

This module defines :class:`AsyncSerialInterface` as a subclass of :class:`fluidlab.interfaces.serial_inter.SerialInterface` and
:class:`pymanip.interface.aiointer.AsyncQueryInterface`.

.. autoclass:: AsyncSerialInterface
   :members:
   :private-members:


"""

import fluidlab.interfaces.serial_inter as fl_serial
from pymanip.interfaces.aiointer import AsyncQueryInterface


class AsyncSerialInterface(fl_serial.SerialInterface, AsyncQueryInterface):
    """This class is an asynchronous extension of :class:`fluidlab.interfaces.serial_inter.SerialInterface`.
    It inherits all its methods from the parent classes.
    """

    pass
